from __future__ import annotations
import typing
from types import CodeType

import sympy
from structlog import get_logger

from . import atoms
from .ode import ODE
from . import sympytools
from ._enum import DeprecatedEnum

logger = get_logger()


def default_printer(
    lhs: sympy.Symbol | sympy.IndexedBase,
    rhs: sympy.Expr,
    use_variable_prefix: bool = False,
) -> str:
    from sympy.codegen.ast import Assignment
    from sympy.printing import pycode

    return pycode(Assignment(lhs, rhs))


class printer_func(typing.Protocol):
    def __call__(
        self,
        lhs: sympy.Symbol | sympy.IndexedBase,
        rhs: sympy.Expr,
        use_variable_prefix: bool = False,
    ) -> str: ...


class scheme_func(typing.Protocol):
    __code__: CodeType

    def __call__(
        self,
        ode: ODE,
        dt: sympy.Symbol,
        name: str = "values",
        printer: printer_func = default_printer,
        remove_unused: bool = False,
    ) -> list[str]: ...


class Scheme(DeprecatedEnum):
    explicit_euler = "explicit_euler"
    generalized_rush_larsen = "generalized_rush_larsen"
    forward_explicit_euler = "forward_explicit_euler", "Use 'explicit_euler' instead"
    forward_generalized_rush_larsen = (
        "forward_generalized_rush_larsen",
        "Use 'generalized_rush_larsen' instead",
    )
    hybrid_rush_larsen = "hybrid_rush_larsen"


def get_scheme(scheme: str) -> scheme_func:
    """Get the scheme function from a string"""
    if scheme in ["forward_euler", "forward_explicit_euler", "euler", "explicit_euler"]:
        func = explicit_euler
    elif scheme in ["forward_generalized_rush_larsen", "generalized_rush_larsen"]:
        func = generalized_rush_larsen
    elif scheme in ["forward_rush_larsen", "rush_larsen", "hybrid_rush_larsen"]:
        func = hybrid_rush_larsen
    else:
        raise ValueError(f"Unknown scheme {scheme}")

    # Replace the name of the function
    func.__code__ = func.__code__.replace(co_name=scheme)
    return func


def list_schemes() -> list[str]:
    """List available schemes"""
    return [s.value for s in Scheme]


def fraction_numerator_is_nonzero(expr):
    """Perform a very cheap check to detect if a fraction is definitely non-zero."""

    if isinstance(expr, sympy.Pow):
        # check if the expression is on the form a**-1
        a, b = expr.args
        if b is sympy.S.NegativeOne:
            return True
        else:
            # we won't do any further checks
            return False
    elif isinstance(expr, sympy.Mul):
        # check if all factors are non-zero
        args = expr.args
        certainly_nonzero_args = []
        potentially_nonzero_args = []
        for e in args:
            if len(e.free_symbols) == 0 and e.is_nonzero:
                certainly_nonzero_args.append(e)
            else:
                potentially_nonzero_args.append(e)

        if len(potentially_nonzero_args) == 0:
            # all factors are certainly nonzero
            return True

        # check all potentially non-zero factors
        for e in potentially_nonzero_args:
            if not fraction_numerator_is_nonzero(e):
                return False
        else:
            return True
    else:
        return False


def explicit_euler(
    ode: ODE,
    dt: sympy.Symbol,
    name: str = "values",
    printer: printer_func = default_printer,
    remove_unused: bool = False,
) -> list[str]:
    r"""Generate forward Euler equations for the ODE

    The forward Euler scheme is given by

    .. math::
        x_{n+1} = x_n + dt f(x_n, t_n)


    Parameters
    ----------
    ode : gotranx.ode.ODE
        The ODE
    dt : sympy.Symbol
        The time step
    name : str, optional
        Name of array to be returned by the scheme, by default "values"
    printer : printer_func, optional
        A code printer, by default default_printer
    remove_unused : bool, optional
        Remove unused variables, by default False

    Returns
    -------
    list[str]
        A list of equations as strings

    """
    logger.debug("Generating explicit Euler scheme")
    eqs = []
    values = sympy.IndexedBase(name, shape=(len(ode.state_derivatives),))
    i = 0
    for x in ode.sorted_assignments(remove_unused=remove_unused):
        eqs.append(printer(x.symbol, x.expr, use_variable_prefix=True))
        if isinstance(x, atoms.StateDerivative):
            eqs.append(
                printer(
                    values[i],
                    x.state.symbol + dt * x.symbol,
                )
            )

            i += 1

    return eqs


def hybrid_rush_larsen(
    ode: ODE,
    dt: sympy.Symbol,
    name: str = "values",
    printer: printer_func = default_printer,
    remove_unused: bool = False,
    delta: float = 1e-8,
    stiff_states: list[str] | None = None,
) -> list[str]:
    r"""Generate the hybrid Rush-Larsen scheme for the ODE

    The hybrid Rush-Larsen scheme follows the standard Rush_Larsen scheme is given by

    .. math::
        x_{n+1} = x_n + \frac{f(x_n, t_n)}{g(x_n, t_n)} \left( e^{g(x_n, t_n) dt} - 1 \right)

    where :math:`g(x_n, t_n)` is the linearization of :math:`f(x_n, t_n)`
    around :math:`x_n`. The difference between the hybrid and the standard
    is that the user can specify which states are stiff, and the RL scheme
    will only be used for these states. If the derivative
    of a state is zero, the scheme falls back to forward Euler.

    We fall back to forward Euler if the derivative is zero.

    Parameters
    ----------
    ode : gotranx.ODE
        The ODE
    dt : sympy.Symbol
        The time step
    name : str, optional
        Name of array to be returned by the scheme, by default "values"
    printer : printer_func, optional
        A code printer, by default default_printer
    remove_unused : bool, optional
        Remove unused variables, by default False
    delta : float, optional
        Tolerance for zero division check, by default 1e-8
    stiff_states : list[str] | None, optional
        Stiff states, by default None. If no stiff states are provided,
        the hybrid rush larsen scheme will be the same as the explicit Euler scheme

    Returns
    -------
    list[str]
        A list of equations as strings

    """
    if stiff_states is None:
        stiff_states = []
    logger.debug("Generating hybrid Rush-Larsen scheme", stiff_states=stiff_states)
    stiff_states_set = set(stiff_states) or set()
    found_stiff_states_set = set()
    eqs = []
    values = sympy.IndexedBase(name, shape=(len(ode.state_derivatives),))
    i = 0
    for x in ode.sorted_assignments(remove_unused=remove_unused):
        eqs.append(printer(x.symbol, x.expr, use_variable_prefix=True))

        if not isinstance(x, atoms.StateDerivative):
            continue

        expr_diff = x.expr.diff(x.state.symbol)
        state_is_stiff = x.state.name in stiff_states_set

        if not state_is_stiff or expr_diff.is_zero:
            # Use forward Euler
            eqs.append(
                printer(
                    values[i],
                    x.state.symbol + dt * x.symbol,
                )
            )
            i += 1
            continue

        found_stiff_states_set.add(x.state.name)
        logger.debug(f"State {x.state.name} is stiff")
        linearized_name = x.name + "_linearized"
        linearized = sympy.Symbol(linearized_name)
        eqs.append(printer(linearized, expr_diff, use_variable_prefix=True))

        need_zero_div_check = not fraction_numerator_is_nonzero(expr_diff)
        if not need_zero_div_check:
            logger.debug(f"{linearized_name} cannot be zero. Skipping zero division check")

        RL_term = x.symbol / linearized * (sympy.exp(linearized * dt) - 1)
        if need_zero_div_check:
            RL_term = sympytools.Conditional(
                abs(linearized) > delta,
                RL_term,
                dt * x.symbol,
            )
        eqs.append(
            printer(
                values[i],
                x.state.symbol + RL_term,
            )
        )
        i += 1
    logger.debug(
        "The following states where marked as stiff but not found in the ODE:",
        extra=stiff_states_set.difference(found_stiff_states_set),
    )
    return eqs


def generalized_rush_larsen(
    ode: ODE,
    dt: sympy.Symbol,
    name: str = "values",
    printer: printer_func = default_printer,
    remove_unused: bool = False,
    delta: float = 1e-8,
) -> list[str]:
    r"""Generate the forward generalized Rush-Larsen scheme for the ODE

    The forward generalized Rush-Larsen scheme is given by

    .. math::
        x_{n+1} = x_n + \frac{f(x_n, t_n)}{g(x_n, t_n)} \left( e^{g(x_n, t_n) dt} - 1 \right)


    where :math:`g(x_n, t_n)` is the linearization of :math:`f(x_n, t_n)` around :math:`x_n`

    We fall back to forward Euler if the derivative is zero.

    Parameters
    ----------
    ode : gotranx.ode.ODE
        The ODE
    dt : sympy.Symbol
        The time step
    name : str, optional
        Name of array to be returned by the scheme, by default "values"
    printer : printer_func, optional
        A code printer, by default default_printer
    remove_unused : bool, optional
        Remove unused variables, by default False
    delta : float, optional
        Tolerance for zero division check, by default 1e-8

    Returns
    -------
    list[str]
        A list of equations as strings
    """
    logger.debug("Generating generalized Rush-Larsen scheme")
    eqs = []
    values = sympy.IndexedBase(name, shape=(len(ode.state_derivatives),))
    i = 0
    for x in ode.sorted_assignments(remove_unused=remove_unused):
        eqs.append(printer(x.symbol, x.expr, use_variable_prefix=True))

        if not isinstance(x, atoms.StateDerivative):
            continue

        expr_diff = x.expr.diff(x.state.symbol)

        if expr_diff.is_zero:
            # Use forward Euler
            eqs.append(
                printer(
                    values[i],
                    x.state.symbol + dt * x.symbol,
                )
            )
            i += 1
            continue

        linearized_name = x.name + "_linearized"
        linearized = sympy.Symbol(linearized_name)
        eqs.append(printer(linearized, expr_diff, use_variable_prefix=True))

        need_zero_div_check = not fraction_numerator_is_nonzero(expr_diff)
        if not need_zero_div_check:
            logger.debug(f"{linearized_name} cannot be zero. Skipping zero division check")

        RL_term = x.symbol / linearized * (sympy.exp(linearized * dt) - 1)
        if need_zero_div_check:
            RL_term = sympytools.Conditional(
                abs(linearized) > delta,
                RL_term,
                dt * x.symbol,
            )
        eqs.append(
            printer(
                values[i],
                x.state.symbol + RL_term,
            )
        )
        i += 1
    return eqs



import typing
from pathlib import Path
import warnings

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated  # type: ignore

import typer

from ..schemes import Scheme, get_scheme
from ..codegen import PythonFormat, CFormat
from . import gotran2c, gotran2py
from . import utils

app = typer.Typer()


def version_callback(show_version: bool):
    """Prints version information."""
    if show_version:
        from .. import __version__, __program_name__

        typer.echo(f"{__program_name__} {__version__}")
        raise typer.Exit()


def license_callback(show_license: bool):
    """Prints license information."""
    if show_license:
        from .. import __license__

        typer.echo(f"{__license__}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version",
    ),
    license: bool = typer.Option(
        None,
        "--license",
        callback=license_callback,
        is_eager=True,
        help="Show license",
    ),
): ...


@app.command()
def convert(
    fname: typing.Optional[Path] = typer.Argument(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    to: str = typer.Option(
        "",
        "--to",
        help="Generate code to another programming language",
    ),
    outname: typing.Optional[str] = typer.Option(
        None,
        "-o",
        "--outname",
        help="Output name",
    ),
    remove_unused: bool = typer.Option(
        False,
        "--remove-unused",
        help="Remove unused variables",
    ),
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version",
    ),
    license: bool = typer.Option(
        None,
        "--license",
        callback=license_callback,
        is_eager=True,
        help="Show license",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Verbose output",
    ),
    scheme: Annotated[
        typing.Optional[typing.List[Scheme]],
        typer.Option(help="Numerical scheme for solving the ODE"),
    ] = None,
    stiff_states: Annotated[
        typing.Optional[typing.List[str]],
        typer.Option("-s", "--stiff-states", help="Stiff states for the hybrid rush larsen scheme"),
    ] = None,
    delta: float = typer.Option(
        1e-8,
        help="Delta value for the rush larsen schemes",
    ),
):
    warnings.warn(
        "convert command is deprecated, use ode2c, ode2py or cellml2ode instead",
        DeprecationWarning,
        stacklevel=1,
    )

    if fname is None:
        return typer.echo("No file specified")

    if to == "":
        # Check if outname is specified
        if outname is None:
            return typer.echo("No output name specified")
        else:
            to = Path(outname).suffix

    if to in {".c", ".h", "c"}:
        gotran2c.main(
            fname=fname,
            suffix=to,
            outname=outname,
            scheme=scheme,
            remove_unused=remove_unused,
            verbose=verbose,
            stiff_states=stiff_states,
            delta=delta,
        )
    if to in {".py", "python", "py"}:
        gotran2py.main(
            fname=fname,
            suffix=to,
            outname=outname,
            scheme=scheme,
            remove_unused=remove_unused,
            verbose=verbose,
            stiff_states=stiff_states,
            delta=delta,
        )

    if to in {".ode"}:
        from .cellml2ode import main as _main

        _main(fname=fname, outname=outname, verbose=verbose)


@app.command()
def cellml2ode(
    fname: typing.Optional[Path] = typer.Argument(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    outname: typing.Optional[str] = typer.Option(
        None,
        "-o",
        "--outname",
        help="Output name",
    ),
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version",
    ),
    license: bool = typer.Option(
        None,
        "--license",
        callback=license_callback,
        is_eager=True,
        help="Show license",
    ),
    config: typing.Optional[Path] = typer.Option(
        None,
        "--config",
        help="Read configuration options from a configuration file",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Verbose output",
    ),
):
    if fname is None:
        return typer.echo("No file specified")

    config_data = utils.read_config(config)
    verbose = config_data.get("verbose", verbose)

    from .cellml2ode import main as _main

    _main(fname=fname, outname=outname, verbose=verbose)


@app.command()
def ode2py(
    fname: typing.Optional[Path] = typer.Argument(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    outname: typing.Optional[str] = typer.Option(
        None,
        "-o",
        "--outname",
        help="Output name",
    ),
    remove_unused: bool = typer.Option(
        False,
        "--remove-unused",
        help="Remove unused variables",
    ),
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version",
    ),
    license: bool = typer.Option(
        None,
        "--license",
        callback=license_callback,
        is_eager=True,
        help="Show license",
    ),
    config: typing.Optional[Path] = typer.Option(
        None,
        "-c",
        "--config",
        help="Read configuration options from a configuration file",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Verbose output",
    ),
    scheme: Annotated[
        typing.List[Scheme],
        typer.Option(help="Numerical scheme for solving the ODE"),
    ] = [],
    stiff_states: Annotated[
        typing.List[str],
        typer.Option("-s", "--stiff-states", help="Stiff states for the hybrid rush larsen scheme"),
    ] = [],
    delta: float = typer.Option(
        1e-8,
        help="Delta value for the rush larsen schemes",
    ),
    format: PythonFormat = typer.Option(
        PythonFormat.black,
        "--format",
        "-f",
        help="Formatter for the output code",
    ),
):
    if fname is None:
        return typer.echo("No file specified")

    config_data = utils.read_config(config)
    verbose = config_data.get("verbose", verbose)
    delta = config_data.get("delta", delta)
    stiff_states = config_data.get("stiff_states", stiff_states)
    scheme = config_data.get("scheme", scheme)
    scheme = utils.validate_scheme(scheme)
    py_config = config_data.get("python", {})
    format = PythonFormat(py_config.get("format", format))

    gotran2py.main(
        fname=fname,
        outname=outname,
        scheme=scheme,
        remove_unused=remove_unused,
        verbose=verbose,
        stiff_states=stiff_states,
        delta=delta,
        format=format,
    )


@app.command()
def ode2c(
    fname: typing.Optional[Path] = typer.Argument(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    to: str = typer.Option(
        ".h",
        "--to",
        help="Generate code to another programming language",
    ),
    outname: typing.Optional[str] = typer.Option(
        None,
        "-o",
        "--outname",
        help="Output name",
    ),
    remove_unused: bool = typer.Option(
        False,
        "--remove-unused",
        help="Remove unused variables",
    ),
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version",
    ),
    license: bool = typer.Option(
        None,
        "--license",
        callback=license_callback,
        is_eager=True,
        help="Show license",
    ),
    config: typing.Optional[Path] = typer.Option(
        None,
        "-c",
        "--config",
        help="Read configuration options from a configuration file",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Verbose output",
    ),
    scheme: Annotated[
        typing.List[Scheme],
        typer.Option(help="Numerical scheme for solving the ODE"),
    ] = [],
    stiff_states: Annotated[
        typing.List[str],
        typer.Option("-s", "--stiff-states", help="Stiff states for the hybrid rush larsen scheme"),
    ] = [],
    delta: float = typer.Option(
        1e-8,
        help="Delta value for the rush larsen schemes",
    ),
    format: CFormat = typer.Option(
        CFormat.clang_format,
        "--format",
        "-f",
        help="Formatter for the output code",
    ),
):
    if fname is None:
        return typer.echo("No file specified")

    config_data = utils.read_config(config)
    verbose = config_data.get("verbose", verbose)
    delta = config_data.get("delta", delta)
    stiff_states = config_data.get("stiff_states", stiff_states)
    scheme = config_data.get("scheme", scheme)
    scheme = utils.validate_scheme(scheme)
    c_config = config_data.get("c", {})
    to = c_config.get("to", to)
    format = CFormat(c_config.get("format", format))

    gotran2c.main(
        fname=fname,
        suffix=to,
        outname=outname,
        scheme=scheme,
        remove_unused=remove_unused,
        verbose=verbose,
        stiff_states=stiff_states,
        delta=delta,
    )


@app.command()
def list_schemes():
    from rich.console import Console
    from rich.table import Table

    table = Table(title="Scheme")

    table.add_column("Name", style="cyan", no_wrap=True)
    table.add_column("Key", style="magenta")
    table.add_column("Extra args", style="green")

    def get_extra_arg_names(f):
        import inspect

        args = inspect.signature(f).parameters.keys()
        return [a for a in args if a not in ["ode", "dt", "name", "printer", "remove_unused"]]

    for scheme in Scheme:
        if scheme in [Scheme.forward_explicit_euler, Scheme.forward_generalized_rush_larsen]:
            # These are deprecated
            continue
        f = get_scheme(scheme.value)
        extra_args = str(get_extra_arg_names(f))
        table.add_row(
            " ".join(map(str.capitalize, scheme.name.split("_"))), scheme.value, extra_args
        )

    console = Console()
    console.print(table)


@app.command()
def inspect(
    fname: typing.Optional[Path] = typer.Argument(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
):
    typer.echo("Hello from inspect")



from __future__ import annotations
from collections import defaultdict

from functools import cached_property
from pathlib import Path
from graphlib import TopologicalSorter
from typing import Iterable
from typing import Sequence
from typing import TypeVar
from typing import cast
from typing import Any
from typing import NamedTuple

import sympy as sp

from . import atoms
from . import exceptions
from .ode_component import BaseComponent, Component

T = TypeVar("T")
U = TypeVar("U", bound=atoms.Assignment)


def check_components(components: Sequence[BaseComponent]):
    """Check if all components are complete

    Parameters
    ----------
    components : Sequence[gotranx.ode_component.BaseComponent]
        The components to check

    Raises
    ------
    exceptions.ComponentNotCompleteError
        If a component is not complete
    """
    for comp in components:
        if not comp.is_complete():
            raise exceptions.ComponentNotCompleteError(
                component_name=comp.name,
                missing_state_derivatives=[state.name for state in comp.states_without_derivatives],
            )


class AllAtoms(NamedTuple):
    symbol_names: list[str]
    symbol_values: dict[str, set[Any]]
    symbols: dict[str, sp.Symbol]
    lookup: dict[str, atoms.Atom]


def gather_atoms(
    components: Sequence[BaseComponent],
) -> AllAtoms:
    """Gather all atoms from a list of components

    Parameters
    ----------
    components : Sequence[gotranx.ode_component.BaseComponent]
        The components to gather atoms from

    Returns
    -------
    AllAtoms
        A named tuple containing all atoms

    """
    symbol_names = []
    symbols = {}
    symbol_values = defaultdict(set)
    lookup: dict[str, atoms.Atom] = {}
    for component in components:
        for p in component.parameters:
            symbol_names.append(p.name)
            symbols[p.name] = p.symbol
            lookup[p.name] = p
            symbol_values[p.name].add(p.value)
        for s in component.states:
            symbol_names.append(s.name)
            symbols[s.name] = s.symbol
            lookup[s.name] = s
            symbol_values[s.name].add(s.value)
        for i in component.intermediates:
            symbol_names.append(i.name)
            symbols[i.name] = i.symbol
            lookup[i.name] = i
            symbol_values[i.name].add(i.expr)
        for st in component.state_derivatives:
            symbol_names.append(st.name)
            symbols[st.name] = st.symbol
            lookup[st.name] = st
    return AllAtoms(symbol_names, symbol_values, symbols, lookup)


def find_duplicates(x: Iterable[T]) -> set[T]:
    """Find duplicates in an iterable.
    Assumes type is hashable

    Parameters
    ----------
    x : Iterable[T]
        The iterable with potential duplicates

    Returns
    -------
    set[T]
        List of duplicate values
    """
    seen = set()
    dupes = []

    for xi in x:
        if xi in seen:
            dupes.append(xi)
        else:
            seen.add(xi)
    return set(dupes)


def add_temporal_state(
    components: Sequence[BaseComponent],
    t: sp.Symbol,
) -> tuple[Component, ...]:
    """Add a temporal state to all components

    Parameters
    ----------
    components : Sequence[gotranx.ode_component.BaseComponent]
        The components to add the temporal state to
    t : sp.Symbol
        The temporal symbol

    Returns
    -------
    tuple[gotranx.ode_component.Component, ...]
        The new components with the temporal state
    """
    new_components = []
    for component in components:
        state_derivatives = set()
        states = set()
        for state_derivative in component.state_derivatives:
            new_state = state_derivative.state.to_TimeDependentState(t)
            state_derivatives.add(
                atoms.StateDerivative(
                    name=state_derivative.name,
                    value=state_derivative.value,
                    components=state_derivative.components,
                    description=state_derivative.description,
                    unit_str=state_derivative.unit_str,
                    unit=state_derivative.unit,
                    state=new_state,
                    symbol=sp.Derivative(new_state.symbol, t),
                ),
            )
            states.add(new_state)

        new_components.append(
            Component(
                name=component.name,
                states=frozenset(states),
                parameters=component.parameters,
                assignments=frozenset(component.intermediates | state_derivatives),
            ),
        )
    return tuple(new_components)


def resolve_expressions(
    components: Sequence[Component],
    symbols: dict[str, sp.Symbol],
) -> tuple[Component, ...]:
    """Resolve all expressions in a list of components

    Parameters
    ----------
    components : Sequence[gotranx.ode_component.Component]
        The components to resolve expressions in
    symbols : dict[str, sp.Symbol]
        The symbols to resolve the expressions with

    Returns
    -------
    tuple[gotranx.ode_component.Component, ...]
        The new components with resolved expressions
    """
    new_components = []
    for component in components:
        assignments = []
        for assignment in component.assignments:
            assignments.append(assignment.resolve_expression(symbols))
        new_components.append(
            Component(
                name=component.name,
                states=component.states,
                parameters=component.parameters,
                assignments=frozenset(assignments),
            ),
        )
    return tuple(new_components)


def make_ode(
    components: Sequence[Component],
    comments: Sequence[atoms.Comment] | None = None,
    name: str = "ODE",
) -> ODE:
    """Create an ODE from a list of components

    Parameters
    ----------
    components : Sequence[gotranx.ode_component.Component]
        The components to create the ODE from
    comments : Sequence[atoms.Comment] | None, optional
        The a list of comments, by default None
    name : str, optional
        Name of the ODE, by default "ODE"

    Returns
    -------
    ODE
        The ODE

    Raises
    ------
    exceptions.DuplicateSymbolError
        If a symbol is duplicated
    """
    check_components(components=components)
    t = sp.Symbol("t")
    # components = add_temporal_state(components, t)
    check_components(components=components)
    _, symbol_values, symbols, lookup = gather_atoms(components=components)
    symbols["time"] = t
    symbols["t"] = t

    if any(x > 1 for x in map(len, symbol_values.values())):
        raise exceptions.DuplicateSymbolError(
            set(k for k, v in symbol_values.items() if len(v) > 1)
        )
    components = resolve_expressions(components=components, symbols=symbols)
    return ODE(components=components, t=t, name=name, comments=comments)


def sort_assignments(
    assignments: Iterable[atoms.Assignment],
    assignments_only: bool = True,
) -> tuple[str, ...]:
    """Sort assignments by dependencies using a topological sorter

    Parameters
    ----------
    assignments : Iterable[atoms.Assignment]
        The assignments to sort
    assignments_only : bool, optional
        If True only include assignments. If False you also include
        states and parameters, by default True.

    Returns
    -------
    tuple[str, ...]
        The sorted assignments

    Raises
    ------
    exceptions.GotranxError
        If an assignment has a None value
    """
    sorter: TopologicalSorter = TopologicalSorter()
    assignment_names = set()
    for assignment in assignments:
        assignment_names.add(assignment.name)
        if assignment.value is None:
            msg = (
                "Unable to sort assignments with None value"
                "Try to save the ODE to an .ode file first and load it again"
            )
            raise exceptions.GotranxError(msg)
        sorter.add(assignment.name, *assignment.value.dependencies)

    static_order = tuple(sorter.static_order())

    if assignments_only:
        static_order = tuple([name for name in static_order if name in assignment_names])

    return static_order


class ODE:
    """A class representing an ODE

    Parameters
    ----------
    components : Sequence[gotranx.ode_component.BaseComponent]
        The components of the ODE
    t : sp.Symbol | None, optional
        Symbol representing time, by default None
    name : str, optional
        Name of the ODE, by default "ODE"
    comments : Sequence[atoms.Comment] | None, optional
        List of comments, by default None

    Raises
    ------
    exceptions.DuplicateSymbolError
        If a symbol is duplicated

    """

    def __init__(
        self,
        components: Sequence[BaseComponent],
        t: sp.Symbol | None = None,
        name: str = "ODE",
        comments: Sequence[atoms.Comment] | None = None,
    ):
        check_components(components)
        _, symbol_values, symbols, lookup = gather_atoms(components=components)

        if any(x > 1 for x in map(len, symbol_values.values())):
            raise exceptions.DuplicateSymbolError(
                set(k for k, v in symbol_values.items() if len(v) > 1)
            )

        t = sp.Symbol("t")
        self.t = t
        symbols["time"] = t

        self._symbols = symbols
        self._lookup = lookup
        self.components = components
        self.name = name
        if comments is None:
            comments = (atoms.Comment(""),)
        self._components = {comp.name: comp for comp in components}

        self.comments = comments
        self.text = " ".join(comment.text for comment in comments)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.name}, "
            f"num_states={self.num_states}, "
            f"num_parameters={self.num_parameters})"
        )

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, ODE):
            return False

        return (
            __o.comments == self.comments
            and __o.components == self.components
            and __o.name == self.name
        )

    def simplify(self) -> ODE:
        """Run sympy's simplify function on all expressions in the ODE"""
        return ODE(
            components=tuple([comp.simplify() for comp in self.components]),
            t=self.t,
            name=self.name,
            comments=self.comments,
        )

    @property
    def states(self) -> tuple[atoms.State, ...]:
        """Get all states in the ODE"""
        states: set[atoms.State] = set()
        for component in self.components:
            states |= component.states
        return tuple(sorted(states, key=lambda x: x.name))

    @property
    def num_states(self) -> int:
        """Get the number of states in the ODE"""
        return len(self.states)

    @property
    def num_components(self) -> int:
        """Get the number of components in the ODE"""
        return len(self.components)

    @cached_property
    def parameters(self) -> tuple[atoms.Parameter, ...]:
        """Get all parameters in the ODE"""
        parameters: set[atoms.Parameter] = set()
        for component in self.components:
            parameters |= component.parameters
        return tuple(sorted(parameters, key=lambda x: x.name))

    @property
    def num_parameters(self) -> int:
        """Get the number of parameters in the ODE"""
        return len(self.parameters)

    @cached_property
    def state_derivatives(self) -> tuple[atoms.StateDerivative, ...]:
        """Get all state derivatives in the ODE sorted by name"""
        state_derivatives: set[atoms.StateDerivative] = set()
        for component in self.components:
            state_derivatives |= component.state_derivatives
        return tuple(sorted(state_derivatives, key=lambda x: x.name))

    @cached_property
    def intermediates(self) -> tuple[atoms.Intermediate, ...]:
        """Get all intermediates in the ODE sorted by name"""
        intermediates: set[atoms.Intermediate] = set()
        for component in self.components:
            intermediates |= component.intermediates
        return tuple(sorted(intermediates, key=lambda x: x.name))

    @property
    def symbols(self) -> dict[str, sp.Symbol]:
        """Get a dictionary of all symbols in the ODE
        with the symbol name as key and the symbol as value
        """
        return self._symbols

    def __sub__(self, other: BaseComponent) -> ODE:
        new_components = [comp for comp in self.components if comp != other]
        return ODE(
            components=new_components,
            t=self.t,
            name=f"{self.name} - {other.name}",
            comments=self.comments,
        )

    def __getitem__(self, name) -> atoms.Atom:
        return self._lookup[name]

    def get_component(self, name: str) -> BaseComponent:
        """Get a component by name

        Parameters
        ----------
        name : str
            Name of the component
        """
        return self._components[name]

    def sorted_assignments(
        self, assignments_only: bool = True, remove_unused: bool = False
    ) -> tuple[atoms.Assignment, ...]:
        """Get the assignments in the ODE sorted by dependencies

        Parameters
        ----------
        assignments_only : bool, optional
            If True only return assignments otherwise you can
            include states and parameters as well, by default True
        remove_unused : bool, optional
            Remove unused variables, by default False

        Returns
        -------
        tuple[atoms.Assignment, ...]
            The sorted assignments
        """
        intermediates = self.intermediates
        if remove_unused:
            deps = self.dependents()
            intermediates = tuple([a for a in intermediates if a.name in deps])

        names = sort_assignments(
            assignments=intermediates + self.state_derivatives,
            assignments_only=assignments_only,
        )
        return tuple([cast(atoms.Assignment, self[name]) for name in names])

    def sorted_state_derivatives(self) -> tuple[atoms.StateDerivative, ...]:
        """Get the state derivatives in the ODE sorted by dependencies"""
        return tuple(s for s in self.sorted_assignments() if isinstance(s, atoms.StateDerivative))

    def sorted_states(self) -> tuple[atoms.State, ...]:
        """Get the states in the ODE sorted by dependencies"""
        return tuple(s.state for s in self.sorted_state_derivatives())

    def save(self, path: Path) -> None:
        """Save the ODE to a file

        Parameters
        ----------
        path : Path
            The path to save the ODE to
        """
        from .save import write_ODE_to_ode_file

        write_ODE_to_ode_file(self, path)

    def dependents(self) -> dict[str, set[str]]:
        """Get a dictionary of dependents for each component"""
        dependencies = defaultdict(set)
        for component in self.components:
            for assignment in component.assignments:
                if assignment.value is None:
                    msg = (
                        "Unable to sort assignments with None value"
                        "Try to save the ODE to an .ode file first and load it again"
                    )
                    raise exceptions.GotranxError(msg)
                for dependency in assignment.value.dependencies:
                    dependencies[dependency].add(assignment.name)

        return dict(dependencies)

    @property
    def missing_variables(self) -> dict[str, int]:
        """Get a dictionary of missing variables for each component

        This is relevant if you have different sub odes where the
        states in one sub ode is a parameter in another sub ode

        """
        symbols = set(self.symbols.keys()) | {"t"}
        variable_names = {var for var in self.dependents() if var not in symbols}
        return {var: i for i, var in enumerate(sorted(variable_names))}



from __future__ import annotations
from pathlib import Path
import logging
import structlog

from ..myokit import cellml_to_gotran

logger = structlog.get_logger()


def main(
    fname: Path,
    outname: str | None = None,
    verbose: bool = False,
) -> None:
    loglevel = logging.DEBUG if verbose else logging.INFO
    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(loglevel),
    )
    assert fname.suffix == ".cellml", f"File {fname} must be a cellml file"
    assert fname.exists(), f"File {fname} does not exist"
    logger.info(f"Converting {fname} to gotran ode file")
    ode = cellml_to_gotran(fname)
    out = fname.with_suffix(".ode") if outname is None else Path(outname)
    ode.save(out)
    logger.info(f"Wrote {out}")



from __future__ import annotations
import typing

from . import c
from . import python


class Template(typing.Protocol):
    @staticmethod
    def state_index(data: dict[str, int]) -> str:
        """The state_index function is a function that returns
        the index of the state with the given name.

        Parameters
        ----------
        data : dict[str, int]
            The data containing the state names and their indexes

        Returns
        -------
        str
            The code for the state_index function
        """

    @staticmethod
    def parameter_index(data: dict[str, int]) -> str:
        """The parameter_index function is a function that returns
        the index of the parameter with the given name.

        Parameters
        ----------
        data : dict[str, int]
            The data containing the parameter names and their indexes

        Returns
        -------
        str
            The code for the parameter_index function
        """

    @staticmethod
    def monitor_index(data: dict[str, int]) -> str:
        """The monitor_index function is a function that returns
        the index of the monitor with the given name.

        A monitor is a variable that is not part of the state or parameter
        but is used to monitor the simulation. For example, in a cardiac
        cell models, different currents across the membrane are examples
        of monitors.

        Parameters
        ----------
        data : dict[str, int]
            The data containing the monitor names and their indexes

        Returns
        -------
        str
            The code for the monitor_index function
        """

    @staticmethod
    def missing_index(data: dict[str, int]) -> str:
        """The missing_index function is a function that returns
        the index of the missing value with the given name.

        A missing value is a variable that is not part of the state or parameter
        but is used in the right-hand side of the ODE. This typically happens
        when the model is not fully defined, and some variables are missing, for
        example when you split an ODE into two sub-models and variables needs
        to be passed between them.

        Parameters
        ----------
        data : dict[str, int]
            The data containing the missing value names and their indexes

        Returns
        -------
        str
            The code for the missing_index function
        """

    @staticmethod
    def init_state_values(
        name: str, state_values: list[float], state_names: list[str], code: str
    ) -> str:
        """The init_state_values function is a function that initializes
        the state values of the model.

        The function typically return an array of the state values, or in
        the case of C code, it takes a pointer to the array of state values
        and initializes the values in place.

        Parameters
        ----------
        name : str
            Name of the return variable
        state_values : list[float]
            Default values for the states
        state_names : list[str]
            The names of the states
        code : str
            Additional code to be inserted into the function

        Returns
        -------
        str
            The code for the init_state_values function
        """

    @staticmethod
    def init_parameter_values(
        name: str, parameter_values: list[float], parameter_names: list[str], code: str
    ) -> str:
        """The init_parameter_values function is a function that initializes
        the parameter values of the model.

        The function typically return an array of the parameter values, or in
        the case of C code, it takes a pointer to the array of parameter values
        and initializes the values in place.

        Parameters
        ----------
        name : str
            Name of the return variable
        parameter_values : list[float]
            Default values for the parameters
        parameter_names : list[str]
            The names of the parameters
        code : str
            Additional code to be inserted into the function

        Returns
        -------
        str
            The code for the init_parameter_values function
        """

    @staticmethod
    def method(
        name: str,
        args: str,
        states: str,
        parameters: str,
        values: str,
        return_name: str | None,
        num_return_values: int,
        shape_info: str,
        values_type: str,
        missing_variables: str,
    ) -> str:
        """The method function is a function that generates a method
        for the model.

        One example of a method is the right-hand side of the ODE, which
        calculates the derivatives of the states. Another example is the
        monitor values, which calculates the values of the monitors.
        All the different numerical schemes are also examples of methods.


        Parameters
        ----------
        name : str
            The name of the method
        args : str
            The arguments of the method
        states : str
            The code for assigning the states
        parameters : str
            The code for assigning the parameters
        values : str
            The code for assigning the values
        return_name : str | None
            The name of the return variable
        num_return_values : int
            The number of return values
        shape_info : str
            The shape information of the return values
        values_type : str
            The type of the values
        missing_variables : str
            The code for handling missing variables

        Returns
        -------
        str
            The code for the method
        """


__all__ = ["c", "python", "Template"]



from __future__ import annotations

import typing
from dataclasses import dataclass
from pathlib import Path

import lark

if typing.TYPE_CHECKING:
    from .atoms import Atom


class GotranxError(Exception):
    pass


@dataclass
class ODEFileNotFound(GotranxError):
    fname: Path


@dataclass
class InvalidODEException(GotranxError):
    text: str
    atoms: tuple[Atom, ...]

    def __str__(self) -> str:
        return f"Invalid ODE: \n{self.text!r}\n with output \n{self.atoms}\n"


@dataclass
class StateNotFoundInComponent(GotranxError):
    state_name: str
    component_name: str

    def __str__(self) -> str:
        return (
            f"State with name {self.state_name!r} "
            f"not found in component {self.component_name!r}"
        )


@dataclass
class ParameterNotFoundInComponent(GotranxError):
    parameter_name: str
    component_name: str

    def __str__(self) -> str:
        return (
            f"Parameter with name {self.parameter_name!r} "
            f"not found in component {self.component_name!r}"
        )


@dataclass
class AssignmentNotFoundInComponent(GotranxError):
    assignment_name: str
    component_name: str

    def __str__(self) -> str:
        return (
            f"Assignment with name {self.assignment_name!r} "
            f"not found in component {self.component_name!r}"
        )


@dataclass
class ComponentNotCompleteError(GotranxError):
    component_name: str
    missing_state_derivatives: list[str]

    def __str__(self) -> str:
        return (
            f"Component {self.component_name!r} is not complete. "
            f"Missing state derivatives for {self.missing_state_derivatives!r}"
        )


@dataclass
class DuplicateSymbolError(GotranxError):
    duplicates: set[str]

    def __str__(self) -> str:
        return (
            f"Found multiple definitions for {self.duplicates!r}. "
            "Please make sure to only define each variable once in order "
            "to avoid ambiguity. Equations should be reorderable."
        )


@dataclass
class UnknownTreeTypeError(GotranxError):
    datatype: str
    atom: str

    def __str__(self) -> str:
        return f"Cannot prase tree data type {self.datatype} for atom {self.atom}"


@dataclass
class InvalidTreeError(GotranxError):
    tree: lark.Tree

    def __str__(self) -> str:
        return f"Invalid tree with data attribute {self.tree.data!r} \n{self.tree.pretty()}"


@dataclass
class MissingSymbolError(GotranxError, KeyError):
    symbol: str
    line_no: int

    def __str__(self) -> str:
        return f"Symbol {self.symbol!r} not found in line {self.line_no}"


@dataclass
class ResolveExpressionError(GotranxError, ValueError):
    name: str

    def __str__(self) -> str:
        return f"Unable to resolve expression for {self.name!r}"



from enum import Enum, EnumMeta

# https://stackoverflow.com/questions/62299740/how-do-i-detect-and-invoke-a-function-when-a-python-enum-member-is-accessed


class OnAccess(EnumMeta):
    """
    runs a user-specified function whenever member is accessed
    """

    #
    def __getattribute__(cls, name):
        obj = super().__getattribute__(name)
        if isinstance(obj, Enum) and obj._on_access:
            obj._on_access()
        return obj

    #
    def __getitem__(cls, name):
        member = super().__getitem__(name)
        if member._on_access:
            member._on_access()
        return member

    #
    def __call__(cls, value, names=None, *, module=None, qualname=None, type=None, start=1):
        obj = super().__call__(
            value, names, module=module, qualname=qualname, type=type, start=start
        )
        if isinstance(obj, Enum) and obj._on_access:
            obj._on_access()
        return obj


class DeprecatedEnum(Enum, metaclass=OnAccess):
    #
    def __new__(cls, value, *args):
        member = object.__new__(cls)
        member._value_ = value
        member._args = args
        member._on_access = member.deprecate if args else None
        return member

    #
    def deprecate(self):
        args = (self.name,) + self._args
        import warnings

        warnings.warn(
            "member %r is deprecated; %s" % args,
            DeprecationWarning,
            stacklevel=3,
        )



# flake8: noqa: E501
import math

import lark
import pytest
import sympy as sp
from gotranx.expressions import build_expression
from gotranx.units import ureg


@pytest.mark.parametrize("expr", ["x=1", "x = 1", "x= 1"])
def test_assignment_single_1(expr, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 1
    assert result[0].name == "x"
    assert result[0].value.tree == lark.Tree(
        "scientific",
        [lark.Token("SCIENTIFIC_NUMBER", "1")],
    )


@pytest.mark.parametrize(
    "expr",
    [
        "x = 1 + 2",
        "x= 1 +2",
        "x = 1+ 2",
    ],
)
def test_assignment_single_2_terms(expr, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 1
    assert result[0].name == "x"
    assert result[0].value.tree == lark.Tree(
        lark.Token("RULE", "expression"),
        [
            lark.Tree(lark.Token("RULE", "scientific"), [lark.Token("SCIENTIFIC_NUMBER", "1")]),
            lark.Token("PLUS", "+"),
            lark.Tree(lark.Token("RULE", "scientific"), [lark.Token("SCIENTIFIC_NUMBER", "2")]),
        ],
    )


def test_assignment_single_3_terms(parser, trans):
    tree = parser.parse("x = 1 * 2 + 3")
    result = trans.transform(tree)
    assert len(result) == 1
    assert result[0].name == "x"
    assert result[0].value.tree == lark.Tree(
        lark.Token("RULE", "expression"),
        [
            lark.Tree(
                lark.Token("RULE", "term"),
                [
                    lark.Tree(
                        lark.Token("RULE", "scientific"),
                        [lark.Token("SCIENTIFIC_NUMBER", "1")],
                    ),
                    lark.Token("STAR", "*"),
                    lark.Tree(
                        lark.Token("RULE", "scientific"),
                        [lark.Token("SCIENTIFIC_NUMBER", "2")],
                    ),
                ],
            ),
            lark.Token("PLUS", "+"),
            lark.Tree(lark.Token("RULE", "scientific"), [lark.Token("SCIENTIFIC_NUMBER", "3")]),
        ],
    )


def test_assignment_single_4_terms(parser, trans):
    tree = parser.parse("x = 1 * 2 + 3 - 4")
    result = trans.transform(tree)
    assert len(result) == 1

    assert result[0].name == "x"
    assert result[0].value.tree == lark.Tree(
        lark.Token("RULE", "expression"),
        [
            lark.Tree(
                lark.Token("RULE", "term"),
                [
                    lark.Tree(
                        lark.Token("RULE", "scientific"),
                        [lark.Token("SCIENTIFIC_NUMBER", "1")],
                    ),
                    lark.Token("STAR", "*"),
                    lark.Tree(
                        lark.Token("RULE", "scientific"),
                        [lark.Token("SCIENTIFIC_NUMBER", "2")],
                    ),
                ],
            ),
            lark.Token("PLUS", "+"),
            lark.Tree(lark.Token("RULE", "scientific"), [lark.Token("SCIENTIFIC_NUMBER", "3")]),
            lark.Token("NEG", "-"),
            lark.Tree(lark.Token("RULE", "scientific"), [lark.Token("SCIENTIFIC_NUMBER", "4")]),
        ],
    )


@pytest.mark.parametrize(
    "expr",
    [
        "x = 1 * 2 + 3 - (4 / 5)",
        "x = 1 * 2 + 3 - 4 / 5",
        "x = (1 * 2) + 3 - (4 / 5)",
    ],
)
def test_assignment_single5(expr, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 1
    assert result[0].name == "x"
    assert result[0].value.tree == lark.Tree(
        lark.Token("RULE", "expression"),
        [
            lark.Tree(
                lark.Token("RULE", "term"),
                [
                    lark.Tree(
                        lark.Token("RULE", "scientific"),
                        [lark.Token("SCIENTIFIC_NUMBER", "1")],
                    ),
                    lark.Token("STAR", "*"),
                    lark.Tree(
                        lark.Token("RULE", "scientific"),
                        [lark.Token("SCIENTIFIC_NUMBER", "2")],
                    ),
                ],
            ),
            lark.Token("PLUS", "+"),
            lark.Tree(lark.Token("RULE", "scientific"), [lark.Token("SCIENTIFIC_NUMBER", "3")]),
            lark.Token("NEG", "-"),
            lark.Tree(
                lark.Token("RULE", "term"),
                [
                    lark.Tree(
                        lark.Token("RULE", "scientific"),
                        [lark.Token("SCIENTIFIC_NUMBER", "4")],
                    ),
                    lark.Token("SLASH", "/"),
                    lark.Tree(
                        lark.Token("RULE", "scientific"),
                        [lark.Token("SCIENTIFIC_NUMBER", "5")],
                    ),
                ],
            ),
        ],
    )


def test_assignment_single_with_names(parser, trans):
    expr = "x = (1 * y) + rho - (z / sigma)"
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 1
    assert result[0].name == "x"
    assert result[0].value.tree == lark.Tree(
        lark.Token("RULE", "expression"),
        [
            lark.Tree(
                lark.Token("RULE", "term"),
                [
                    lark.Tree(
                        lark.Token("RULE", "scientific"),
                        [lark.Token("SCIENTIFIC_NUMBER", "1")],
                    ),
                    lark.Token("STAR", "*"),
                    lark.Tree(lark.Token("RULE", "variable"), [lark.Token("VARIABLE", "y")]),
                ],
            ),
            lark.Token("PLUS", "+"),
            lark.Tree(lark.Token("RULE", "variable"), [lark.Token("VARIABLE", "rho")]),
            lark.Token("NEG", "-"),
            lark.Tree(
                lark.Token("RULE", "term"),
                [
                    lark.Tree(lark.Token("RULE", "variable"), [lark.Token("VARIABLE", "z")]),
                    lark.Token("SLASH", "/"),
                    lark.Tree(
                        lark.Token("RULE", "variable"),
                        [lark.Token("VARIABLE", "sigma")],
                    ),
                ],
            ),
        ],
    )


@pytest.mark.parametrize("expr", ["x=1\ny=2", "x = 1 \n y =2", "x= 1\n y = 2"])
def test_assignment_double(expr, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 2
    assert result[0].name == "x"
    assert result[0].value.tree == lark.Tree(
        "scientific",
        [lark.Token("SCIENTIFIC_NUMBER", "1")],
    )
    assert result[1].name == "y"
    assert result[1].value.tree == lark.Tree(
        "scientific",
        [lark.Token("SCIENTIFIC_NUMBER", "2")],
    )


def test_expressions_with_name_only(parser, trans):
    expr = """
    expressions("My Component")
    x = 1
    y = 2
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    components = result.components

    assert len(components) == 1
    component = components[0]
    assert component.name == "My Component"

    assignments = component.assignments
    assert len(assignments) == 2

    x = component.find_assignment("x")
    assert x.name == "x"
    assert x.value.tree == lark.Tree(
        "scientific",
        [lark.Token("SCIENTIFIC_NUMBER", "1")],
    )

    y = component.find_assignment("y")
    assert y.name == "y"
    assert y.value.tree == lark.Tree(
        "scientific",
        [lark.Token("SCIENTIFIC_NUMBER", "2")],
    )


def test_expressions_with_two_components(parser, trans):
    expr = """
    expressions("My Component", "My second component")
    x = 1
    y = 2
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)

    components = result.components

    assert len(components) == 2
    for name, component in zip(["My Component", "My second component"], components):
        assert component.name == name
        assignments = component.assignments
        assert len(assignments) == 2

        x = component.find_assignment("x")
        assert x.components == ("My Component", "My second component")
        assert x.name == "x"
        assert x.value.tree == lark.Tree(
            "scientific",
            [lark.Token("SCIENTIFIC_NUMBER", "1")],
        )

        y = component.find_assignment("y")
        assert x.components == ("My Component", "My second component")
        assert y.name == "y"
        assert y.value.tree == lark.Tree(
            "scientific",
            [lark.Token("SCIENTIFIC_NUMBER", "2")],
        )


def test_expressions_with_name_and_unit(parser, trans):
    expr = """
    expressions("My Component", "My second component")
    x = 1  # mV
    y = 2  # mol
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)

    tree = parser.parse(expr)
    result = trans.transform(tree)

    components = result.components

    assert len(components) == 2
    for name, component in zip(["My Component", "My second component"], components):
        assert component.name == name
        assignments = component.assignments
        assert len(assignments) == 2

        x = component.find_assignment("x")
        assert x.comment is None
        assert x.components == ("My Component", "My second component")
        assert x.name == "x"
        assert x.value.tree == lark.Tree(
            "scientific",
            [lark.Token("SCIENTIFIC_NUMBER", "1")],
        )
        assert x.unit == ureg.Unit("mV")

        y = component.find_assignment("y")
        assert y.components == ("My Component", "My second component")
        assert y.name == "y"
        assert y.value.tree == lark.Tree(
            "scientific",
            [lark.Token("SCIENTIFIC_NUMBER", "2")],
        )
        assert y.unit == ureg.Unit("mol")
        assert y.comment is None


def test_expressions_with_invalid_unit_becomes_comment(parser, trans):
    expr = """
    expressions("My Component", "My second component")
    x = 1  # Some variable that we want to define
    y = 2  # 1 + 1
    z = 2  # 3.14
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)

    components = result.components

    assert len(components) == 2
    for name, component in zip(["My Component", "My second component"], components):
        assert component.name == name
        assignments = component.assignments
        assert len(assignments) == 3

        x = component.find_assignment("x")
        assert x.components == ("My Component", "My second component")
        assert x.name == "x"
        assert x.value.tree == lark.Tree(
            "scientific",
            [lark.Token("SCIENTIFIC_NUMBER", "1")],
        )
        assert x.unit is None
        assert x.comment.text == "Some variable that we want to define"

        y = component.find_assignment("y")
        assert y.components == ("My Component", "My second component")
        assert y.name == "y"
        assert y.value.tree == lark.Tree(
            "scientific",
            [lark.Token("SCIENTIFIC_NUMBER", "2")],
        )
        assert y.unit is None
        assert y.comment.text == "1 + 1"

        z = component.find_assignment("z")
        assert x.components == ("My Component", "My second component")
        assert z.name == "z"
        assert z.value.tree == lark.Tree(
            "scientific",
            [lark.Token("SCIENTIFIC_NUMBER", "2")],
        )
        assert z.unit is None
        assert z.comment.text == "3.14"


@pytest.mark.parametrize(
    "expr, subs, expected",
    [
        ("y = log(x * 2)", {"x": 1}, math.log(2.0)),
        ("y = log(x + log(2))", {"x": 1}, math.log(1 + math.log(2.0))),
        ("y = exp(x * 2)", {"x": 1}, math.exp(2.0)),
        ("y = exp(x + log(2) - 1 )", {"x": 1}, 2),
        ("y = sin(x)", {"x": math.pi / 3}, math.sqrt(3) / 2),
        ("y = cos(x)", {"x": math.pi / 3}, 0.5),
        ("y = tan(x)", {"x": math.pi / 3}, math.sqrt(3)),
        ("y = asin(x)", {"x": math.sqrt(3) / 2}, math.pi / 3),
        ("y = acos(x)", {"x": 0.5}, math.pi / 3),
        ("y = atan(x)", {"x": math.sqrt(3)}, math.pi / 3),
        ("y = sqrt(x)", {"x": 4}, 2),
        ("y = Abs(x)", {"x": -2}, 2),
    ],
)
def test_expression_functions(expr, subs, expected, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    symbols = {name: sp.Symbol(name) for name in subs}
    sympy_expr = build_expression(result[0].value.tree, symbols=symbols)
    assert math.isclose(sympy_expr.subs(subs), expected)


@pytest.mark.parametrize(
    "expr, subs, expected",
    [
        ("\n x = 1\n y = x**2", {"x": 1}, 1),
        ("\n x = 1\n y = x**2", {"x": 2}, 4),
        ("\n x = 1\n y = (1 + x) ** 2", {"x": 2}, 9),
        ("\n x = 1\n y = ((2 + x) / 2) ** (1 + x)", {"x": 2}, 8),
    ],
)
def test_power_expressions(expr, subs, expected, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    symbols = {name: sp.Symbol(name) for name in subs}
    y = result.components[0].find_assignment("y")
    sympy_expr = build_expression(y.value.tree, symbols=symbols)
    assert math.isclose(sympy_expr.subs(subs), expected)


@pytest.mark.parametrize(
    "expr, expected",
    [
        ("\n x = pi", math.pi),
        ("\n x = 3 * pi", 3 * math.pi),
    ],
)
def test_expressions_constants(expr, expected, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    x = result.components[0].find_assignment("x")
    sympy_expr = build_expression(x.value.tree)
    assert math.isclose(sympy_expr, expected)


@pytest.mark.parametrize(
    "expr, expected",
    [
        ("\n x = 1e3", 1000),
        ("\n x = 1e-3", 0.001),
        ("\n x = 1E2", 100),
        ("\n x = 1E-2", 0.01),
        ("\n x = 2e3", 2000),
        ("\n x = 2e-3", 0.002),
        ("\n x = 2E2", 200),
        ("\n x = 2E-2", 0.02),
    ],
)
def test_expressions_scientific_notation(expr, expected, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    x = result.components[0].find_assignment("x")
    sympy_expr = build_expression(x.value.tree)
    assert math.isclose(sympy_expr, expected)


@pytest.mark.parametrize(
    "expr, unit",
    [
        ("x = 1", None),
        ("x = 1 # mV", ureg.Unit("mV")),
        ("x = 1 # pA*pF**-1", ureg.Unit("pA*pF**-1")),
    ],
)
def test_assignment_with_unit(expr, unit, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    # Since expression don't start with a newline we get
    # an Assignment object
    x = result[0]
    assert x.unit == unit


@pytest.mark.parametrize(
    "expr, subs, expected",
    [
        (
            "alpha_h = Conditional(Lt(V, -40), 0.057*exp(-(V + 80)/6.8), +0) # ms**-1",
            {"V": 0},
            0,
        ),
        (
            "alpha_h = Conditional(Lt(V, -40), 0.057*exp(-(V + 80)/6.8), 0) # ms**-1",
            {"V": -80},
            0.057,
        ),
        (
            "alpha_h = Conditional(Gt(V, -40), 0.057*exp(-(V + 80)/6.8), 0) # ms**-1",
            {"V": 0},
            0.057 * math.exp(-80 / 6.8),
        ),
        (
            "alpha_h = Conditional(Gt(V, -40), 0.057*exp(-(V + 80)/6.8), 0) # ms**-1",
            {"V": -80},
            0,
        ),
    ],
)
def test_lt_gt_conditional(expr, subs, expected, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    symbols = {name: sp.Symbol(name) for name in subs}
    alpha_h = result[0]
    sympy_expr = build_expression(alpha_h.value.tree, symbols=symbols)

    assert math.isclose(sympy_expr.subs(subs), expected)


@pytest.mark.parametrize(
    "subs, expected",
    [
        (
            {
                "stim_period": 100,
                "time": 0,
                "stim_amplitude": 42,
                "stim_start": 1,
                "stim_duration": 10,
            },
            0,
        ),
        (
            {
                "stim_period": 100,
                "time": 2,
                "stim_amplitude": 42,
                "stim_start": 1,
                "stim_duration": 10,
            },
            -42,
        ),
        (
            {
                "stim_period": 100,
                "time": 12,
                "stim_amplitude": 42,
                "stim_start": 1,
                "stim_duration": 10,
            },
            0,
        ),
        (
            {
                "stim_period": 100,
                "time": 102,
                "stim_amplitude": 42,
                "stim_start": 1,
                "stim_duration": 10,
            },
            -42,
        ),
    ],
)
def test_stimulus_current(subs, expected, parser, trans):
    expr = """
    i_Stim = Conditional(And(Ge(time - floor(time/stim_period)*stim_period, stim_start),Le(time - floor(time/stim_period)*stim_period, stim_start + stim_duration)),-stim_amplitude, 0) # pA*pF**-1
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    i_Stim = result.components[0].find_assignment("i_Stim")
    symbols = {name: sp.Symbol(name) for name in i_Stim.value.dependencies}
    sympy_expr = build_expression(i_Stim.value.tree, symbols=symbols)

    assert math.isclose(sympy_expr.subs(subs), expected)


def test_stimulus_current_with_trailing_comma(parser, trans):
    subs = {
        "stim_period": 100,
        "time": 0,
        "stim_amplitude": 42,
        "stim_start": 1,
        "stim_duration": 10,
    }
    expected = 0
    expr = """
    i_Stim = Conditional(And(Ge(time - floor(time/stim_period)*stim_period, stim_start),Le(time - floor(time/stim_period)*stim_period, stim_start + stim_duration),),-stim_amplitude, 0) # pA*pF**-1
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    i_Stim = result.components[0].find_assignment("i_Stim")
    symbols = {name: sp.Symbol(name) for name in i_Stim.value.dependencies}
    sympy_expr = build_expression(i_Stim.value.tree, symbols=symbols)

    assert math.isclose(sympy_expr.subs(subs), expected)


@pytest.mark.parametrize(
    "subs, expected",
    [
        (
            {
                "time": 0,
                "stim_amplitude": 42,
                "stim_start": 1,
                "stim_duration": 10,
            },
            math.exp(-1 / 0.2) * 42,
        ),
        (
            {
                "time": 2,
                "stim_amplitude": 42,
                "stim_start": 1,
                "stim_duration": 10,
            },
            42 - math.exp(-1 / 0.2) * 42,
        ),
        (
            {
                "time": 12,
                "stim_amplitude": 42,
                "stim_start": 1,
                "stim_duration": 10,
            },
            math.exp(-1 / 0.2) * 42,
        ),
    ],
)
def test_stimulus_current_continuous(subs, expected, parser, trans):
    expr = """
    i_Stim = stim_amplitude * ContinuousConditional(Ge(time, stim_start), 1, 0, 0.2) * ContinuousConditional(Le(time, stim_start + stim_duration), 1, 0, 0.2)
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    i_Stim = result.components[0].find_assignment("i_Stim")
    symbols = {name: sp.Symbol(name) for name in i_Stim.value.dependencies}
    sympy_expr = build_expression(i_Stim.value.tree, symbols=symbols)
    assert math.isclose(sympy_expr.subs(subs), expected, rel_tol=1e-2)



import pytest
import gotranx


@pytest.fixture
def main_ode(trans, parser):
    expr = """
    parameters("Main component",
    sigma=ScalarParam(12.0, description="Some description"),
    rho=21.0
    )

    parameters("Z component",
    beta=2.4
    )

    states("Main component", x=1.0, y=2.0)

    states("Z component", z=3.05)

    expressions("Main component")
    rhoz = rho - z
    dy_dt = x*rhoz - y # millivolt
    dx_dt = sigma*(-x + y)

    expressions("Z component")
    betaz = beta*z
    dz_dt = -betaz + x*y
    """

    tree = parser.parse(expr)
    return gotranx.ode.make_ode(*trans.transform(tree), name="lorentz")


def test_component_to_ode(main_ode):
    z_comp = main_ode.get_component("Z component")
    z_ode = z_comp.to_ode()
    assert z_ode.missing_variables == {"x": 0, "y": 1}
    assert z_ode.num_states == 1
    assert z_ode.num_parameters == 1
    assert z_ode.parameters[0].name == "beta"
    assert z_ode.states[0].name == "z"
    assert z_ode.name == "Z component"

    remaining_ode = main_ode - z_comp
    assert remaining_ode.missing_variables == {"z": 0}
    assert remaining_ode.num_states == 2
    assert remaining_ode.num_parameters == 2
    assert remaining_ode.parameters[0].name == "rho"
    assert remaining_ode.parameters[1].name == "sigma"
    assert remaining_ode.states[0].name == "x"
    assert remaining_ode.states[1].name == "y"
    assert remaining_ode.name == "lorentz - Z component"


@pytest.fixture
def z_ode_codegen(main_ode):
    z_comp = main_ode.get_component("Z component")
    z_ode = z_comp.to_ode()
    return gotranx.codegen.PythonCodeGenerator(z_ode)


def test_codegen_component_ode_missing_index(z_ode_codegen):
    assert z_ode_codegen.missing_index() == (
        'missing = {"x": 0, "y": 1}'
        "\n"
        "\n"
        "\ndef missing_index(name: str) -> int:"
        '\n    """Return the index of the missing with the given name'
        "\n"
        "\n    Arguments"
        "\n    ---------"
        "\n    name : str"
        "\n        The name of the missing"
        "\n"
        "\n    Returns"
        "\n    -------"
        "\n    int"
        "\n        The index of the missing"
        "\n"
        "\n    Raises"
        "\n    ------"
        "\n    KeyError"
        "\n        If the name is not a valid missing"
        '\n    """'
        "\n"
        "\n    return missing[name]"
        "\n"
    )


def test_codegen_component_ode_rhs(z_ode_codegen):
    assert z_ode_codegen.rhs() == (
        "def rhs(t, states, parameters, missing_variables):"
        "\n"
        "\n    # Assign states"
        "\n    z = states[0]"
        "\n"
        "\n    # Assign parameters"
        "\n    beta = parameters[0]"
        "\n"
        "\n    # Assign missing variables"
        "\n    x = missing_variables[0]"
        "\n    y = missing_variables[1]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    betaz = beta * z"
        "\n    dz_dt = -betaz + x * y"
        "\n    values[0] = dz_dt"
        "\n"
        "\n    return values"
        "\n"
    )


def test_codegen_component_ode_monitor(z_ode_codegen):
    assert z_ode_codegen.monitor_values() == (
        "def monitor_values(t, states, parameters, missing_variables):"
        "\n"
        "\n    # Assign states"
        "\n    z = states[0]"
        "\n"
        "\n    # Assign parameters"
        "\n    beta = parameters[0]"
        "\n"
        "\n    # Assign missing variables"
        "\n    x = missing_variables[0]"
        "\n    y = missing_variables[1]"
        "\n"
        "\n    # Assign expressions"
        "\n    shape = 2 if len(states.shape) == 1 else (2, states.shape[1])"
        "\n    values = numpy.zeros(shape)"
        "\n    betaz = beta * z"
        "\n    values[0] = betaz"
        "\n    dz_dt = -betaz + x * y"
        "\n    values[1] = dz_dt"
        "\n"
        "\n    return values"
        "\n"
    )


def test_codegen_component_ode_fe(z_ode_codegen):
    assert z_ode_codegen.scheme(gotranx.get_scheme("forward_euler")) == (
        "def forward_euler(states, t, dt, parameters, missing_variables):"
        "\n"
        "\n    # Assign states"
        "\n    z = states[0]"
        "\n"
        "\n    # Assign parameters"
        "\n    beta = parameters[0]"
        "\n"
        "\n    # Assign missing variables"
        "\n    x = missing_variables[0]"
        "\n    y = missing_variables[1]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    betaz = beta * z"
        "\n    dz_dt = -betaz + x * y"
        "\n    values[0] = dt * dz_dt + z"
        "\n"
        "\n    return values"
        "\n"
    )


@pytest.fixture
def remaining_ode_codegen(main_ode):
    z_comp = main_ode.get_component("Z component")
    remaining_ode = main_ode - z_comp
    return gotranx.codegen.PythonCodeGenerator(remaining_ode)


def test_codegen_remaining_ode_missing_index(remaining_ode_codegen):
    assert remaining_ode_codegen.missing_index() == (
        'missing = {"z": 0}'
        "\n"
        "\n"
        "\ndef missing_index(name: str) -> int:"
        '\n    """Return the index of the missing with the given name'
        "\n"
        "\n    Arguments"
        "\n    ---------"
        "\n    name : str"
        "\n        The name of the missing"
        "\n"
        "\n    Returns"
        "\n    -------"
        "\n    int"
        "\n        The index of the missing"
        "\n"
        "\n    Raises"
        "\n    ------"
        "\n    KeyError"
        "\n        If the name is not a valid missing"
        '\n    """'
        "\n"
        "\n    return missing[name]"
        "\n"
    )


def test_codegen_remaining_ode_rhs(remaining_ode_codegen):
    assert remaining_ode_codegen.rhs() == (
        "def rhs(t, states, parameters, missing_variables):"
        "\n"
        "\n    # Assign states"
        "\n    x = states[0]"
        "\n    y = states[1]"
        "\n"
        "\n    # Assign parameters"
        "\n    rho = parameters[0]"
        "\n    sigma = parameters[1]"
        "\n"
        "\n    # Assign missing variables"
        "\n    z = missing_variables[0]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    rhoz = rho - z"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    values[0] = dx_dt"
        "\n    dy_dt = rhoz * x - y"
        "\n    values[1] = dy_dt"
        "\n"
        "\n    return values"
        "\n"
    )


def test_codegen_remaining_ode_generate_missing_values(remaining_ode_codegen):
    assert remaining_ode_codegen.missing_values({"x": 0, "rhoz": 1}) == (
        "def missing_values(t, states, parameters, missing_variables):"
        "\n"
        "\n    # Assign states"
        "\n    x = states[0]"
        "\n    y = states[1]"
        "\n"
        "\n    # Assign parameters"
        "\n    rho = parameters[0]"
        "\n    sigma = parameters[1]"
        "\n"
        "\n    # Assign missing variables"
        "\n    z = missing_variables[0]"
        "\n"
        "\n    # Assign expressions"
        "\n    shape = 2 if len(states.shape) == 1 else (2, states.shape[1])"
        "\n    values = numpy.zeros(shape)"
        "\n    values[0] = x"
        "\n    rhoz = rho - z"
        "\n    values[1] = rhoz"
        "\n"
        "\n    return values"
        "\n"
    )



from pathlib import Path
import subprocess
import sys
import shutil
import pytest
import os

here = Path(__file__).parent


@pytest.fixture
def env():
    env = os.environ.copy()
    env["MPLBACKEND"] = "Agg"
    return env


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python3.9 or higher")
@pytest.mark.parametrize(
    "example",
    (pytest.param(f, id=f.name) for f in (here / ".." / "examples").iterdir() if f.is_dir()),
)
def test_examples(example, tmpdir, env):
    if not (example / "main.py").exists():
        return

    shutil.copytree(example, tmpdir / example.name)

    ret = subprocess.run(
        [sys.executable, "main.py"], cwd=tmpdir / example.name, check=True, env=env
    )
    assert ret.returncode == 0



from __future__ import annotations
import enum
import typing
import structlog
from sympy.printing.c import C99CodePrinter
from sympy.codegen.ast import Assignment
import sympy

from ..ode import ODE
from .. import templates
from .base import CodeGenerator, Func, RHSArgument, SchemeArgument

logger = structlog.get_logger()


class Format(str, enum.Enum):
    clang_format = "clang-format"
    none = "none"


def get_formatter(format: Format) -> typing.Callable[[str], str]:
    if format == Format.none:
        return lambda x: x
    elif format == Format.clang_format:
        try:
            import clang_format_docs
        except ImportError:
            logger.warning("Cannot apply clang-format, please install 'clang-format-docs'")
            return lambda x: x
        else:
            return clang_format_docs.clang_format_str
    else:
        raise ValueError(f"Unknown format: {format}")


def bool_to_int(expr: str) -> str:
    return expr.replace("false", "0").replace("true", "1")


class GotranCCodePrinter(C99CodePrinter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._settings["contract"] = False

    def _print_Float(self, flt):
        return self._print(str(float(flt)))

    def _print_Piecewise(self, expr):
        if isinstance(expr.args[0][0], Assignment):
            result = []
            lhs = super()._print(expr.args[0][0].lhs)
            result.append(f"{super()._print(lhs)} = ")
            all_lsh_equal = True
            for arg in expr.args:
                result.append(f"({super()._print(arg[1])}) ? ")
                result.append(f"{super()._print(arg[0].rhs)}")
                result.append(" : ")
                all_lsh_equal = all_lsh_equal and super()._print(arg[0].lhs) == lhs

            assert all_lsh_equal, "All assignments in Piecewise must have the same lhs"

            if super()._print(arg[1]) == "true":
                result = result[:-3]
                result.append(f"{super()._print(arg[0].rhs)}")
            else:
                raise ValueError("Last condition in Piecewise must be True")

            result.append(";")
            value = "".join(result)
        else:
            value = bool_to_int(super()._print_Piecewise(expr))

        return value


class CCodeGenerator(CodeGenerator):
    variable_prefix = "const double "

    def __init__(
        self, ode: ODE, format: Format = Format.clang_format, remove_unused: bool = False
    ) -> None:
        super().__init__(ode, remove_unused=remove_unused)
        self._printer = GotranCCodePrinter()
        setattr(self, "_formatter", get_formatter(format=format))

    @property
    def printer(self):
        return self._printer

    @property
    def template(self):
        return templates.c

    def imports(self) -> str:
        return self._format(
            "\n".join(
                [
                    "#include <math.h>",
                    "#include <string.h>\n",
                ]
            )
        )

    def _rhs_arguments(
        self, order: RHSArgument | str = RHSArgument.stp, const_states: bool = True
    ) -> Func:
        value = RHSArgument.get_value(order)
        states_prefix = "const " if const_states else ""
        argument_dict = {
            "s": states_prefix + "double *__restrict states",
            "t": "const double t",
            "p": "const double *__restrict parameters",
        }
        argument_list = [argument_dict[v] for v in value] + ["double* values"]
        states = sympy.IndexedBase("states", shape=(self.ode.num_states,))
        parameters = sympy.IndexedBase("parameters", shape=(self.ode.num_parameters,))
        values = sympy.IndexedBase("values", shape=(self.ode.num_states,))

        return Func(
            arguments=argument_list,
            states=states,
            parameters=parameters,
            values=values,
            values_type="",
        )

    def _scheme_arguments(
        self,
        order: SchemeArgument | str = SchemeArgument.stdp,
        const_states: bool = True,
    ) -> Func:
        value = SchemeArgument.get_value(order)
        states_prefix = "const " if const_states else ""
        argument_dict = {
            "s": states_prefix + "double *__restrict states",
            "t": "const double t",
            "d": "const double dt",
            "p": "const double *__restrict parameters",
        }
        argument_list = [argument_dict[v] for v in value] + ["double* values"]
        states = sympy.IndexedBase("states", shape=(self.ode.num_states,))
        parameters = sympy.IndexedBase("parameters", shape=(self.ode.num_parameters,))
        values = sympy.IndexedBase("values", shape=(self.ode.num_states,))

        return Func(
            arguments=argument_list,
            states=states,
            parameters=parameters,
            values=values,
            values_type="",
        )



from __future__ import annotations
from pathlib import Path
import logging
import structlog

from ..codegen.python import PythonCodeGenerator, get_formatter, Format
from ..load import load_ode
from ..schemes import Scheme
from ..ode import ODE

from .utils import add_schemes

logger = structlog.get_logger()


def get_code(
    ode: ODE,
    scheme: list[Scheme] | None = None,
    format: Format = Format.black,
    remove_unused: bool = False,
    missing_values: dict[str, int] | None = None,
    delta: float = 1e-8,
    stiff_states: list[str] | None = None,
) -> str:
    """Generate the Python code for the ODE

    Parameters
    ----------
    ode : gotranx.ode.ODE
        The ODE
    scheme : list[Scheme] | None, optional
        Optional numerical scheme, by default None
    format : gotranx.codegen.python.Format, optional
        The formatter, by default gotranx.codegen.python.Format.black
    remove_unused : bool, optional
        Remove unused variables, by default False
    missing_values : dict[str, int] | None, optional
        Missing values, by default None
    delta : float, optional
        Delta value for the rush larsen schemes, by default 1e-8
    stiff_states : list[str] | None, optional
        Stiff states, by default None. Only applicable for
        the hybrid rush larsen scheme

    Returns
    -------
    str
        The Python code
    """
    codegen = PythonCodeGenerator(
        ode,
        format=Format.none,
        remove_unused=remove_unused,
    )
    formatter = get_formatter(format=format)
    if missing_values is not None:
        _missing_values = codegen.missing_values(missing_values)
    else:
        _missing_values = ""

    comp = [
        codegen.imports(),
        codegen.parameter_index(),
        codegen.state_index(),
        codegen.monitor_index(),
        codegen.missing_index(),
        codegen.initial_parameter_values(),
        codegen.initial_state_values(),
        codegen.rhs(),
        codegen.monitor_values(),
        _missing_values,
    ] + add_schemes(
        codegen,
        scheme=scheme,
        delta=delta,
        stiff_states=stiff_states,
    )
    code = codegen._format("\n".join(comp))

    if format != Format.none:
        # Run the formatter only once
        logger.debug("Applying formatter", format=format)
        code = formatter(code)
    return code


def main(
    fname: Path,
    outname: str | None = None,
    format: Format = Format.black,
    scheme: list[Scheme] | None = None,
    remove_unused: bool = False,
    verbose: bool = True,
    stiff_states: list[str] | None = None,
    delta: float = 1e-8,
    suffix: str = ".py",
) -> None:
    loglevel = logging.DEBUG if verbose else logging.INFO
    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(loglevel),
    )

    ode = load_ode(fname)
    code = get_code(
        ode,
        scheme=scheme,
        format=format,
        remove_unused=remove_unused,
        stiff_states=stiff_states,
        delta=delta,
    )
    out = fname if outname is None else Path(outname)
    out_name = out.with_suffix(suffix=suffix)
    out_name.write_text(code)
    logger.info(f"Wrote {out_name}")



import pytest
from gotranx.expressions import build_expression
from gotranx import codegen
import sympy as sp


@pytest.mark.parametrize(
    "expr, expected",
    [
        ("x = Conditional(Eq(time, 0), 1, 0)", (1, 0, 0)),
        ("x = Conditional(Gt(time, 0), 1, 0)", (0, 1, 0)),
        ("x = Conditional(Ge(time, 0), 1, 0)", (1, 1, 0)),
        ("x = Conditional(Lt(time, 0), 1, 0)", (0, 0, 1)),
        ("x = Conditional(Le(time, 0), 1, 0)", (1, 0, 1)),
        ("x = Conditional(Not(Eq(time, 0)), 1, 0)", (0, 1, 1)),
        ("x = Conditional(Not(Gt(time, 0)), 1, 0)", (1, 0, 1)),
        ("x = Conditional(Not(Ge(time, 0)), 1, 0)", (0, 0, 1)),
        ("x = Conditional(Not(Lt(time, 0)), 1, 0)", (1, 1, 0)),
        ("x = Conditional(Not(Le(time, 0)), 1, 0)", (0, 1, 0)),
    ],
)
def test_Conditional_expr(expr, expected, parser, trans):
    time = sp.Symbol("time")
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 1
    sympy_expr = build_expression(result[0].value.tree, symbols={"time": time})
    assert sympy_expr.subs({"time": 0}) == expected[0]
    assert sympy_expr.subs({"time": 1}) == expected[1]
    assert sympy_expr.subs({"time": -1}) == expected[2]


@pytest.mark.parametrize(
    "Printer, expected",
    [
        (codegen.BaseGotranODECodePrinter, "Conditional(Lt(t, 0), a, b)"),
        (codegen.GotranPythonCodePrinter, "numpy.where((t < 0), a, b)"),
        (codegen.GotranCCodePrinter, "((t < 0) ? (\n   a\n)\n: (\n   b\n))"),
    ],
)
def test_single_Conditional_from_sympy(Printer, expected):
    t = sp.Symbol("t")
    a = sp.Symbol("a")
    b = sp.Symbol("b")

    expr = sp.Piecewise((a, sp.Lt(t, 0)), (b, True))
    result = Printer().doprint(expr)
    assert result == expected


@pytest.mark.parametrize(
    "Printer, expected",
    [
        (
            codegen.BaseGotranODECodePrinter,
            "Conditional(Lt(t, 0), a, Conditional(Gt(t, 0), b, c))",
        ),
        (
            codegen.GotranPythonCodePrinter,
            "numpy.where((t < 0), a, numpy.where((t > 0), b, c))",
        ),
        (
            codegen.GotranCCodePrinter,
            "((t < 0) ? (\n   a\n)\n: ((t > 0) ? (\n   b\n)\n: (\n   c\n)))",
        ),
    ],
)
def test_nested_Conditional_from_sympy(Printer, expected):
    t = sp.Symbol("t")
    a = sp.Symbol("a")
    b = sp.Symbol("b")
    c = sp.Symbol("c")

    expr = sp.Piecewise((a, t < 0), (b, t > 0), (c, True))
    result = Printer().doprint(expr)

    assert result == expected


def test_And_Or_from_sympy():
    t = sp.Symbol("t")
    expr = sp.Or(sp.Lt(t, 0), sp.And(sp.Gt(t, 0), sp.Lt(t, 1)))
    result = codegen.BaseGotranODECodePrinter().doprint(expr)

    assert result == "Or(Lt(t, 0), And(Gt(t, 0), Lt(t, 1)))"



"""Python template for code generation"""

from __future__ import annotations
from textwrap import dedent, indent
import functools
from structlog import get_logger

logger = get_logger()


def acc(all_values: str, next_value: str = "# ") -> str:
    """Accumulate values in a string

    Arguments
    ---------
    all_values : str
        The accumulated values
    next_value : str
        The next value to add

    Returns
    -------
    str
        The accumulated values
    """

    if len(all_values.split("\n")[-1]) + len(next_value) > 60:
        return all_values + "\n#" + next_value
    else:
        return all_values + ", " + next_value


def _index(data: dict[str, int], name: str) -> str:
    """Helper function to generate index function for a given data

    Arguments
    ---------
    data : dict[str, int]
        The data
    name : str
        The name of the function

    Returns
    -------
    str
        The index function
    """
    return dedent(
        f'''
{name} = {repr(data)}


def {name}_index(name: str) -> int:
    """Return the index of the {name} with the given name

    Arguments
    ---------
    name : str
        The name of the {name}

    Returns
    -------
    int
        The index of the {name}

    Raises
    ------
    KeyError
        If the name is not a valid {name}
    """

    return {name}[name]
''',
    )


def state_index(data: dict[str, int]) -> str:
    """The state_index function is a function that returns
    the index of the state with the given name.

    Parameters
    ----------
    data : dict[str, int]
        The data containing the state names and their indexes

    Returns
    -------
    str
        The code for the state_index function
    """

    logger.debug(f"Generating state_index with {len(data)} values")
    return _index(data, "state")


def parameter_index(data: dict[str, int]) -> str:
    """The parameter_index function is a function that returns
    the index of the parameter with the given name.

    Parameters
    ----------
    data : dict[str, int]
        The data containing the parameter names and their indexes

    Returns
    -------
    str
        The code for the parameter_index function
    """

    logger.debug(f"Generating parameter_index with {len(data)} values")
    return _index(data, "parameter")


def monitor_index(data: dict[str, int]) -> str:
    """The monitor_index function is a function that returns
    the index of the monitor with the given name.

    A monitor is a variable that is not part of the state or parameter
    but is used to monitor the simulation. For example, in a cardiac
    cell models, different currents across the membrane are examples
    of monitors.

    Parameters
    ----------
    data : dict[str, int]
        The data containing the monitor names and their indexes

    Returns
    -------
    str
        The code for the monitor_index function
    """
    logger.debug(f"Generating monitored_index with {len(data)} values")
    return _index(data, "monitor")


def missing_index(data: dict[str, int]) -> str:
    """The missing_index function is a function that returns
    the index of the missing value with the given name.

    A missing value is a variable that is not part of the state or parameter
    but is used in the right-hand side of the ODE. This typically happens
    when the model is not fully defined, and some variables are missing, for
    example when you split an ODE into two sub-models and variables needs
    to be passed between them.

    Parameters
    ----------
    data : dict[str, int]
        The data containing the missing value names and their indexes

    Returns
    -------
    str
        The code for the missing_index function
    """
    logger.debug(f"Generating missing_index with {len(data)} values")
    return _index(data, "missing")


def init_state_values(name, state_names, state_values, code):
    """The init_state_values function is a function that initializes
    the state values of the model.

    The function typically return an array of the state values, or in
    the case of C code, it takes a pointer to the array of state values
    and initializes the values in place.

    Parameters
    ----------
    name : str
        Name of the return variable
    state_values : list[float]
        Default values for the states
    state_names : list[str]
        The names of the states
    code : str
        Additional code to be inserted into the function (not used in this template)

    Returns
    -------
    str
        The code for the init_state_values function
    """
    logger.debug(f"Generating init_state_values with {len(state_values)} values")
    values_comment = indent(
        "#" + functools.reduce(acc, [f"{n}={v}" for n, v in zip(state_names, state_values)]),
        "    ",
    )

    values = ", ".join(map(str, state_values))
    return dedent(
        f'''
def init_state_values(**values):
    """Initialize state values
    """
{values_comment}

    {name} = numpy.array([{values}], dtype=numpy.float64)

    for key, value in values.items():
        {name}[state_index(key)] = value

    return {name}
''',
    )


def init_parameter_values(name, parameter_names, parameter_values, code):
    """The init_parameter_values function is a function that initializes
    the parameter values of the model.

    The function typically return an array of the parameter values, or in
    the case of C code, it takes a pointer to the array of parameter values
    and initializes the values in place.

    Parameters
    ----------
    name : str
        Name of the return variable
    parameter_values : list[float]
        Default values for the parameters
    parameter_names : list[str]
        The names of the parameters
    code : str
        Additional code to be inserted into the function (not used in this template)

    Returns
    -------
    str
        The code for the init_parameter_values function
    """
    logger.debug(f"Generating init_parameter_values with {len(parameter_values)} values")
    values_comment = indent(
        "#"
        + functools.reduce(acc, [f"{n}={v}" for n, v in zip(parameter_names, parameter_values)]),
        "    ",
    )

    values = ", ".join(map(str, parameter_values))
    return dedent(
        f'''
def init_parameter_values(**values):
    """Initialize parameter values
    """
{values_comment}

    {name} = numpy.array([{values}], dtype=numpy.float64)

    for key, value in values.items():
        {name}[parameter_index(key)] = value

    return {name}
''',
    )


def method(
    name,
    args,
    states,
    parameters,
    values,
    return_name: str,
    num_return_values: int,
    nan_to_num: bool = False,
    values_type: str = "numpy.zeros_like(states, dtype=numpy.float64)",
    shape_info: str = "",
    missing_variables: str = "",
    **kwargs,
):
    """The method function is a function that generates a method
    for the model.

    One example of a method is the right-hand side of the ODE, which
    calculates the derivatives of the states. Another example is the
    monitor values, which calculates the values of the monitors.
    All the different numerical schemes are also examples of methods.


    Parameters
    ----------
    name : str
        The name of the method
    args : str
        The arguments of the method
    states : str
        The code for assigning the states
    parameters : str
        The code for assigning the parameters
    values : str
        The code for assigning the values
    return_name : str | None
        The name of the return variable
    num_return_values : int
        The number of return values
    shape_info : str
        The shape information of the return values
    values_type : str
        The type of the values
    missing_variables : str
        The code for handling missing variables
    nan_to_num : bool
        If True, the return values are passed through numpy.nan_to_num
        with nan=0.0

    Returns
    -------
    str
        The code for the method
    """

    logger.debug(f"Generating method '{name}', with {num_return_values} return values.")
    if len(kwargs) > 0:
        logger.debug(f"Unused kwargs: {kwargs}")

    indent_missing_variables = indent(missing_variables, "    ")
    indent_states = indent(states, "    ")
    indent_parameters = indent(parameters, "    ")
    indent_values = indent(values, "    ")
    if nan_to_num:
        indent_return = indent(
            f"return numpy.nan_to_num({return_name}, nan=0.0)",
            "    ",
        )
    else:
        indent_return = indent(f"return {return_name}", "    ")
    return dedent(
        f"""
def {name}({args}):

    # Assign states
{indent_states}

    # Assign parameters
{indent_parameters}
{indent_missing_variables}
    # Assign expressions
    {shape_info}
    {return_name} = {values_type}
{indent_values}

{indent_return}
""",
    )



from __future__ import annotations

from collections import defaultdict
from typing import NamedTuple
from typing import Type
from typing import TypeVar

import lark

from . import atoms
from . import exceptions
from . import ode_component
from . import units

T = TypeVar("T", atoms.State, atoms.Parameter)


class LarkODE(NamedTuple):
    components: tuple[ode_component.Component, ...]
    comments: tuple[atoms.Comment, ...]


def remove_quotes(s: str) -> str:
    """Remove quotes from a string

    Parameters
    ----------
    s : str
        The string

    Returns
    -------
    str
        The string without quotes
    """
    return s.replace("'", "").replace('"', "")


def get_unit_and_comment_from_assignment(
    s: lark.Tree,
) -> tuple[str | None, atoms.Comment | None]:
    """Get the unit and comment from an assignment

    Parameters
    ----------
    s : lark.Tree
        The tree

    Returns
    -------
    tuple[str | None, atoms.Comment | None]
        The unit and comment
    """
    # If there are less than 3 children, there is no unit
    if len(s.children) >= 3:
        # The third child is the potential unit
        potential_unit = s.children[2]
        # If it's a comment, it's not a unit
        if isinstance(potential_unit, atoms.Comment):
            try:
                # Try to parse the unit
                unit = units.ureg(potential_unit.text)
            except (units.pint.UndefinedUnitError, AttributeError):
                # Not a proper unit so it's a comment
                return None, atoms.Comment(potential_unit.text)
            else:
                if isinstance(unit, units.pint.Quantity):
                    # It's a proper unit
                    return potential_unit.text, None

                # It is something else, so it's a comment
                return None, atoms.Comment(potential_unit.text)

        else:
            return None, None

    return None, None


def find_assignments(
    s: lark.Tree | lark.lexer.Token | None,
    components: tuple[str, ...],
) -> list[atoms.Assignment]:
    """Find assignments in a tree

    Parameters
    ----------
    s : lark.Tree | lark.lexer.Token | None
        The tree
    components : tuple[str, ...]
        List of components

    Returns
    -------
    list[atoms.Assignment]
        List of assignments
    """
    if isinstance(s, lark.Tree):
        unit_str, comment = get_unit_and_comment_from_assignment(s)
        return [
            atoms.Assignment(
                name=str(s.children[0]),
                value=atoms.Expression(tree=s.children[1]),
                components=tuple(components),
                unit_str=unit_str,
                comment=comment,
            ),
        ]

    return []


def find_components(
    s: list[None | lark.tree.Tree | lark.lexer.Token],
) -> tuple[int, tuple[str, ...]]:
    """Find components in a list

    Parameters
    ----------
    s : list[None  |  lark.tree.Tree  |  lark.lexer.Token]
        The list

    Returns
    -------
    tuple[int, tuple[str, ...]]
        The index and the components
    """
    i = 0
    components = []

    while (
        isinstance(s[i], lark.lexer.Token)
        and hasattr(s[i], "type")
        and s[i].type == "COMPONENT_NAME"  # type:ignore
    ):
        components.append(remove_quotes(str(s[i])))
        i += 1

    # If no components are found, add an empty string
    if len(components) == 0:
        components.append("")
    return i, tuple(components)


def lark_list_to_parameters(
    s: list[None | lark.tree.Tree | lark.lexer.Token],
    cls: Type[T],
) -> tuple[T, ...]:
    """Convert a list of lark trees to parameters

    Parameters
    ----------
    s : list[None  |  lark.tree.Tree  |  lark.lexer.Token]
        The list
    cls : Type[T]
        The class of the parameters

    Returns
    -------
    tuple[T, ...]
        The parameters
    """
    i, components = find_components(s)
    return tuple(
        [
            tree2parameter(p, components=components, cls=cls)
            for p in s[i:]
            if isinstance(p, lark.Tree)
        ],
    )


def tree2parameter(
    s: lark.Tree,
    components: tuple[str, ...],
    cls: Type[T],
) -> T:
    """Convert a lark tree to a parameter

    Parameters
    ----------
    s : lark.Tree
        The tree
    components : tuple[str, ...]
        The components
    cls : Type[T]
        The class of the parameter

    Returns
    -------
    T
        The parameter

    Raises
    ------
    exceptions.UnknownTreeTypeError
        If the tree type is unknown
    """
    from .expressions import build_expression

    if s.data == "param":
        return cls(
            name=str(s.children[0]),
            value=build_expression(s.children[1]),
            components=tuple(components),
        )
    elif s.data == "scalarparam":
        unit = None
        if s.children[2] is not None:
            unit = remove_quotes(str(s.children[2]))

        desc = None
        if s.children[3] is not None:
            desc = remove_quotes(str(s.children[3]))

        return cls(
            name=str(s.children[0]),
            value=build_expression(s.children[1]),
            components=tuple(components),
            unit_str=unit,
            description=desc,
        )
    else:
        raise exceptions.UnknownTreeTypeError(datatype=s.data, atom="Parameter")


class TreeToODE(lark.Transformer):
    """Transform a lark tree to an ODE

    See https://lark-parser.readthedocs.io/en/latest/recipes.html
    for more information on how to use lark transformers.
    """

    def _call_userfunc(self, tree, new_children=None):  # pragma: nocover
        # Assumes tree is already transformed
        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            try:
                wrapper = getattr(f, "visit_wrapper", None)
                if wrapper is not None:
                    return f.visit_wrapper(f, tree.data, children, tree.meta)
                else:
                    return f(children)
            except lark.GrammarError:
                raise

    def _call_userfunc_token(self, token):  # pragma: nocover
        try:
            f = getattr(self, token.type)
        except AttributeError:
            return self.__default_token__(token)
        else:
            try:
                return f(token)
            except lark.GrammarError:
                raise

    def comment(self, s):
        """Convert a comment to an atoms.Comment"""
        return atoms.Comment(" ".join(map(str.lstrip, map(lambda x: x.lstrip("#"), map(str, s)))))

    def states(self, s: list[None | lark.tree.Tree | lark.lexer.Token]) -> tuple[atoms.State, ...]:
        """Convert a list of states to atoms.State"""
        return lark_list_to_parameters(s, cls=atoms.State)

    def parameters(
        self, s: list[None | lark.tree.Tree | lark.lexer.Token]
    ) -> tuple[atoms.Parameter, ...]:
        """Convert a list of parameters to atoms.Parameter"""
        return lark_list_to_parameters(s, cls=atoms.Parameter)

    def expressions(self, s) -> tuple[atoms.Assignment, ...]:
        """Convert a list of expressions to atoms.Assignment"""
        i, components = find_components(s)

        assignments = []

        for si in s[i:]:
            assignments.extend(find_assignments(si, components=components))

        return tuple(assignments)

    def ode(self, s) -> LarkODE:
        """Convert a lark tree to an ODE

        Parameters
        ----------
        s : list[Any]
            List of objects that could by states, parameters or assignments

        Returns
        -------
        LarkODE
            A tuple of components and comments
        """
        # FIXME: Could use Enum here
        mapping = {
            atoms.Parameter: "parameters",
            atoms.Assignment: "assignments",
            atoms.State: "states",
        }

        components: dict[str, dict[str, set[atoms.Atom]]] = defaultdict(
            lambda: {atom: set() for atom in mapping.values()},
        )

        # breakpoint()

        comments = []
        for line in s:  # Each line in the block
            if isinstance(line, atoms.Comment):
                comments.append(line)
                continue
            if isinstance(line, str):
                assert line.strip() == "", f"Invalid line {line!r}"
                # Skip empty lines
                continue

            for atom in line:  # State, Parameters or Assignment
                for component in atom.components:
                    components[component][mapping[type(atom)]].add(atom)

        # Make sets frozen
        frozen_components: dict[str, dict[str, frozenset[atoms.Atom]]] = {}
        for component_name, component_values in components.items():
            frozen_components[component_name] = {}
            for atom_name, atom_values in component_values.items():
                frozen_components[component_name][atom_name] = frozenset(atom_values)

        # FIXME: Need to somehow tell the type checker that each of the inner dictionaries
        # are actually of the correct type.
        return LarkODE(
            components=tuple(
                [
                    ode_component.Component(name=key, **value)  # type: ignore
                    for key, value in frozen_components.items()
                ],
            ),
            comments=tuple(comments),
        )



from . import base
from . import c
from . import python
from . import ode

from .c import CCodeGenerator, GotranCCodePrinter, Format as CFormat
from .python import PythonCodeGenerator, GotranPythonCodePrinter, Format as PythonFormat
from .base import CodeGenerator, Func, RHSArgument, SchemeArgument
from .ode import GotranODECodePrinter, BaseGotranODECodePrinter

__all__ = [
    "base",
    "c",
    "CCodeGenerator",
    "CodeGenerator",
    "Func",
    "RHSArgument",
    "SchemeArgument",
    "python",
    "PythonCodeGenerator",
    "GotranPythonCodePrinter",
    "ode",
    "GotranODECodePrinter",
    "BaseGotranODECodePrinter",
    "GotranCCodePrinter",
    "CFormat",
    "PythonFormat",
]



from __future__ import annotations

from pathlib import Path

from structlog import get_logger

from .codegen.ode import GotranODECodePrinter
from .ode import ODE

logger = get_logger()


def write_ODE_to_ode_file(ode: ODE, path: Path) -> None:
    """Write an ODE to a file

    Parameters
    ----------
    ode : gotranx.ode.ODE
        The ODE
    path : Path
        The path to write the ODE to
    """
    # Just make sure it is a Path object
    printer = GotranODECodePrinter(ode)
    path = Path(path)
    text = []
    text.append(printer.print_comments())
    text.append(printer.print_states())
    text.append(printer.print_parameters())
    text.append(printer.print_assignments())
    path.write_text("".join(text))
    logger.info(f"Wrote {path}")



from textwrap import dedent
import gotranx
import pytest
from pathlib import Path
from typer.testing import CliRunner

here = Path(__file__).parent.absolute()
runner = CliRunner(mix_stderr=False)


@pytest.fixture(scope="session")
def config_file(tmp_path_factory):
    text = dedent(
        """
        [tool.gotranx]
        scheme = ["hybrid_rush_larsen"]
        stiff_states = ["x", "y"]
        verbose = true
        delta = 1e-3

        [tool.gotranx.python]
        format = "ruff"

        [tool.gotranx.c]
        format = "none"
        to = ".c"
        """
    )
    fname = tmp_path_factory.mktemp("data") / "pyproject.toml"
    fname.write_text(text)
    yield fname
    fname.unlink()


@pytest.fixture(scope="module")
def all_schemes():
    args = []
    for scheme in gotranx.schemes.Scheme:
        args.append("--scheme")
        args.append(scheme.value)
    return args


@pytest.fixture(scope="module")
def odefile(tmp_path_factory):
    expr = """
    parameters(a=0)
    parameters("My component",
    sigma=ScalarParam(12.0, description="Some description"),
    rho=21.0,
    beta=2.4
    )
    states("My component", x=1.0, y=2.0,z=3.05)

    expressions("My component")
    dy_dt = x*(rho - z) - y # millivolt
    dx_dt = sigma*(-x + y)
    dz_dt = -beta*z + x*y
    """
    fname = tmp_path_factory.mktemp("data") / "lorentz.ode"
    fname.write_text(expr)
    yield fname
    fname.unlink()


def test_cli_version():
    result = runner.invoke(gotranx.cli.app, "--version")
    assert result.exit_code == 0
    assert "gotranx" in result.stdout


def test_cli_license():
    result = runner.invoke(gotranx.cli.app, "--license")
    assert result.exit_code == 0
    assert "MIT" in result.stdout


def test_gotran2py_old(odefile, all_schemes):
    outfile = odefile.with_suffix(".py")

    result = runner.invoke(
        gotranx.cli.app, ["convert", str(odefile), "--to", ".py", "-o", str(outfile)] + all_schemes
    )
    assert result.exit_code == 0
    assert "lorentz.py" in result.stdout
    assert "lorentz" in result.stdout
    assert outfile.is_file()

    code = outfile.read_text()
    for scheme in gotranx.schemes.Scheme:
        assert scheme.value in code
    assert "rhs" in code
    assert "init_state_values" in code
    assert "init_parameter_values" in code
    assert "monitor_index" in code
    assert "state_index" in code
    assert "parameter_index" in code

    outfile.unlink()


def test_gotran2c_old(odefile, all_schemes):
    outfile = odefile.with_suffix(".h")
    result = runner.invoke(
        gotranx.cli.app, ["convert", str(odefile), "--to", ".h", "-o", str(outfile)] + all_schemes
    )
    assert result.exit_code == 0

    assert "lorentz.h" in result.stdout
    assert "lorentz" in result.stdout
    assert outfile.is_file()

    code = outfile.read_text()
    for scheme in gotranx.schemes.Scheme:
        assert scheme.value in code
    assert "rhs" in code
    assert "init_state_values" in code
    assert "init_parameter_values" in code
    assert "monitor_index" in code
    assert "state_index" in code
    assert "parameter_index" in code

    outfile.with_suffix(".h").unlink()


def test_cellml2ode_old():
    cellmlfile = here / "cellml_files" / "noble_1962.cellml"
    out_odefile = cellmlfile.with_suffix(".ode")
    result = runner.invoke(
        gotranx.cli.app,
        ["convert", str(cellmlfile), "-o", cellmlfile.with_suffix(".ode")],
    )
    assert result.exit_code == 0

    assert f"Wrote {out_odefile}" in result.stdout
    assert out_odefile.is_file()
    # Check that we can load the file
    ode = gotranx.load_ode(out_odefile)
    assert ode is not None
    out_odefile.unlink()


def test_cellml2ode():
    cellmlfile = here / "cellml_files" / "noble_1962.cellml"
    out_odefile = cellmlfile.with_suffix(".ode")
    result = runner.invoke(
        gotranx.cli.app,
        ["cellml2ode", str(cellmlfile), "-o", cellmlfile.with_suffix(".ode")],
    )
    assert result.exit_code == 0

    assert f"Wrote {out_odefile}" in result.stdout
    assert out_odefile.is_file()
    # Check that we can load the file
    ode = gotranx.load_ode(out_odefile)
    assert ode is not None
    out_odefile.unlink()


@pytest.mark.parametrize("format", gotranx.codegen.PythonFormat)
def test_gotran2py(format, odefile, all_schemes):
    outfile = odefile.with_suffix(".py")

    stiff_states = ["-s", "x", "-s", "y", "-s", "w"]

    result = runner.invoke(
        gotranx.cli.app,
        ["ode2py", str(odefile), "-v", "-o", str(outfile), "-f", format.value]
        + all_schemes
        + stiff_states,
    )
    assert result.exit_code == 0
    assert "lorentz.py" in result.stdout
    assert "lorentz" in result.stdout
    if format != gotranx.codegen.PythonFormat.none:
        assert "Applying formatter" in result.stdout
        assert format.value in result.stdout

    assert outfile.is_file()

    code = outfile.read_text()
    for scheme in gotranx.schemes.Scheme:
        assert scheme.value in code
    assert "rhs" in code
    assert "init_state_values" in code
    assert "init_parameter_values" in code
    assert "monitor_index" in code
    assert "state_index" in code
    assert "parameter_index" in code

    outfile.unlink()


@pytest.mark.parametrize("format", gotranx.codegen.CFormat)
@pytest.mark.parametrize("suffix", [".h", ".c"])
def test_gotran2c(format, suffix, odefile, all_schemes):
    outfile = odefile.with_suffix(suffix)
    result = runner.invoke(
        gotranx.cli.app,
        ["ode2c", str(odefile), "-v", "--to", suffix, "-f", format.value, "-o", str(outfile)]
        + all_schemes,
    )
    assert result.exit_code == 0

    assert f"lorentz{suffix}" in result.stdout
    assert "lorentz" in result.stdout
    assert outfile.is_file()

    code = outfile.read_text()
    for scheme in gotranx.schemes.Scheme:
        assert scheme.value in code
    assert "rhs" in code
    assert "init_state_values" in code
    assert "init_parameter_values" in code
    assert "monitor_index" in code
    assert "state_index" in code
    assert "parameter_index" in code

    outfile.with_suffix(suffix).unlink()


def test_ode2py_config_file(odefile, config_file):
    outfile = odefile.with_suffix(".py")
    result = runner.invoke(
        gotranx.cli.app,
        ["ode2py", str(odefile), "-c", str(config_file), "-o", str(outfile)],
    )
    assert result.exit_code == 0
    assert "lorentz.py" in result.stdout
    assert "lorentz" in result.stdout
    assert "stiff_states=['x', 'y']" in result.stdout
    assert "Applying formatter" in result.stdout
    assert "ruff" in result.stdout
    assert outfile.is_file()
    code = outfile.read_text()
    assert "hybrid_rush_larsen" in code
    outfile.unlink()


def test_ode2c_config_file(odefile, config_file):
    outfile = odefile.with_suffix(".c")
    result = runner.invoke(
        gotranx.cli.app,
        ["ode2c", str(odefile), "-c", str(config_file), "-o", str(outfile)],
    )
    assert result.exit_code == 0
    assert "lorentz.c" in result.stdout
    assert "lorentz" in result.stdout
    assert "stiff_states=['x', 'y']" in result.stdout
    assert outfile.is_file()
    code = outfile.read_text()
    assert "hybrid_rush_larsen" in code
    outfile.unlink()



from __future__ import annotations
from typing import Any
from pathlib import Path

import typer

from ..codegen import CodeGenerator
from ..schemes import Scheme, get_scheme


def add_schemes(
    codegen: CodeGenerator,
    scheme: list[Scheme] | None = None,
    delta: float = 1e-8,
    stiff_states: list[str] | None = None,
) -> list[str]:
    comp = []
    if scheme is not None:
        for s in scheme:
            kwargs: dict[str, Any] = {}
            if "rush_larsen" in s.value:
                kwargs["delta"] = delta
            if s.value == "hybrid_rush_larsen":
                kwargs["stiff_states"] = stiff_states

            comp.append(codegen.scheme(get_scheme(s.value), **kwargs))
    return comp


def find_pyproject_toml_config() -> Path | None:
    """Find the pyproject.toml file."""
    from black.files import find_pyproject_toml

    path = find_pyproject_toml((str(Path.cwd()),))
    if path is None:
        return None
    return Path(path)


def read_config(path: Path | None) -> dict[str, Any]:
    """Read the configuration file."""

    # If no path is given, try to find the pyproject.toml file
    if path is None:
        path = find_pyproject_toml_config()

    # Return empty dict if no path is found
    if path is None:
        return {}

    # Try to read the configuration file
    try:
        # First try to use tomllib which is part of stdlib
        import tomllib as toml
    except ImportError:
        # If tomllib is not available, try to use toml
        try:
            import toml  # type: ignore
        except ImportError:
            typer.echo("Please install 'tomllib' or 'toml' to read configuration files")
            return {}

    try:
        config = toml.loads(Path(path).read_text())
    except Exception:
        typer.echo(f"Could not read configuration file {path}")
        return {}
    else:
        return config.get("tool", {}).get("gotranx", {})


def validate_scheme(scheme: list[Scheme] | list[str]) -> list[Scheme]:
    lst = []
    for s in scheme:
        if isinstance(s, str):
            lst.append(Scheme(s))
        else:
            lst.append(s)
    return lst



from __future__ import annotations

import attr
import lark
import pint
import sympy as sp
from structlog import get_logger

from .expressions import build_expression
from .units import ureg
from . import exceptions

logger = get_logger()


def _set_symbol(instance, name: str) -> None:
    """Helper function to set the symbol attribute of a frozen instance

    Parameters
    ----------
    instance : Any
        The instance to set the symbol attribute on
    name : str
        The name of the symbol
    """
    object.__setattr__(
        instance,
        "symbol",
        sp.Symbol(
            name=name,
            real=True,
            imaginary=False,
            commutative=True,
            finite=True,
        ),
    )


def unit_from_string(unit_str: str | None) -> pint.Unit | None:
    """Create a pint unit from a string

    Parameters
    ----------
    unit_str : str | None
        The string representation of the unit

    Returns
    -------
    pint.Unit | None
        It the unit is valid, return the pint unit, else None
    """
    if unit_str is not None:
        try:
            unit = ureg.Unit(unit_str)
        except pint.UndefinedUnitError:
            logger.warning(f"Undefined unit {unit_str!r}")
            unit = None
        except ValueError:
            try:
                unit = ureg.Unit(unit_str.split(" ")[0])
            except Exception:
                logger.warning(f"Invalid unit {unit_str!r}")
                unit = None
    else:
        unit = None
    return unit


def _set_unit(instance, unit_str: str) -> None:
    """Helper function to set the unit attribute of a frozen instance
    by parsing a string representation of the unit

    Parameters
    ----------
    instance : Any
        The instance to set the unit attribute on
    unit_str : str
        The string representation of the unit
    """
    object.__setattr__(instance, "unit", unit_from_string(unit_str))


@attr.s(frozen=True, slots=True)
class Comment:
    """A comment is a string that is not parsed by the parser. It is
    used to add human readable information to the model.
    """

    text: str = attr.ib()


@attr.s(frozen=True, kw_only=True, slots=True)
class Atom:
    """Base class for atoms"""

    name: str = attr.ib()
    value: float | Expression | sp.core.Number | None = attr.ib()
    components: tuple[str, ...] = attr.ib(default=("",))
    description: str | None = attr.ib(None)
    symbol: sp.Symbol = attr.ib(None)
    unit_str: str | None = attr.ib(None, repr=False)
    unit: pint.Unit | None = attr.ib(None)

    def __attrs_post_init__(self):
        if self.unit is None:
            _set_unit(self, unit_str=self.unit_str)
        if self.symbol is None:
            _set_symbol(self, name=self.name)


@attr.s(frozen=True, kw_only=True, slots=True)
class Parameter(Atom):
    """A Parameter is a constant scalar value"""

    value: float | sp.core.Number = attr.ib()


@attr.s(frozen=True, kw_only=True, slots=True)
class State(Atom):
    """A State is a variable that also has a
    corresponding state derivative.
    """

    value: float | sp.core.Number = attr.ib()

    def to_TimeDependentState(self, t: sp.Symbol) -> "TimeDependentState":
        return TimeDependentState(
            name=self.name,
            value=self.value,
            symbol=sp.Function(self.symbol)(t),
            components=self.components,
            description=self.description,
            unit_str=self.unit_str,
            unit=self.unit,
        )


@attr.s(frozen=True, kw_only=True, slots=True)
class TimeDependentState(State):
    """A TimeDependentState is a State, where the symbol
    is a sympy Function instead of a pure Symbol.
    """

    symbol: sp.Function = attr.ib()


@attr.s(frozen=True, kw_only=True, slots=True)
class Expression:
    """An Expression is a group of variables (i.e
    states, parameters or other expressions) combined
    with binary operations (i.e +, -, * etc)
    An Expression is typically a right hand side of
    an assignment."""

    tree: lark.Tree = attr.ib(cmp=False)  # Different trees can give same expression
    dependencies: frozenset[str] = attr.ib(init=False)

    def __attrs_post_init__(self):
        object.__setattr__(self, "dependencies", self._find_dependencies())

    def _find_dependencies(self) -> frozenset[str]:
        deps = set()

        for tree in self.tree.iter_subtrees():
            if tree.data == "variable":
                deps.add(str(tree.children[0]))
        return frozenset(deps)

    def resolve(self, symbols: dict[str, sp.Symbol]):
        return build_expression(self.tree, symbols=symbols)


@attr.s(frozen=True, kw_only=True, slots=True)
class Assignment(Atom):
    """Assignments are object of the form `name = value`."""

    value: Expression | None = attr.ib()
    expr: sp.Expr = attr.ib(sp.S.Zero)
    comment: Comment | None = attr.ib(None)

    def resolve_expression(self, symbols: dict[str, sp.Symbol]) -> Assignment:
        """Resolve the expression of the assignment by
        building the sympy expression from the expression tree

        Parameters
        ----------
        symbols : dict[str, sp.Symbol]
            A dictionary of all symbols in the model

        Returns
        -------
        Assignment
            A new Assignment object with the resolved expression

        Raises
        ------
        exceptions.ResolveExpressionError
            If the expression is not set
        """
        if self.value is None:
            raise exceptions.ResolveExpressionError(name=self.name)
        expr = self.value.resolve(symbols)
        return type(self)(
            name=self.name,
            value=self.value,
            components=self.components,
            unit_str=self.unit_str,
            unit=self.unit,
            expr=expr,
            symbol=self.symbol,
            description=self.description,
            comment=self.comment,
        )

    def to_intermediate(self) -> "Intermediate":
        """Convert the Assignment to an Intermediate"""
        return Intermediate(
            name=self.name,
            value=self.value,
            components=self.components,
            unit_str=self.unit_str,
            unit=self.unit,
            expr=self.expr,
            description=self.description,
            symbol=self.symbol,
            comment=self.comment,
        )

    def to_state_derivative(self, state: State) -> "StateDerivative":
        """Convert the Assignment to a StateDerivative

        Parameters
        ----------
        state : State
            The associated state
        """
        return StateDerivative(
            name=self.name,
            value=self.value,
            components=self.components,
            unit_str=self.unit_str,
            unit=self.unit,
            state=state,
            expr=self.expr,
            description=self.description,
            symbol=self.symbol,
            comment=self.comment,
        )

    def simplify(self) -> "Assignment":
        """Simplify the expression of the assignment using sympy's
        simplify function. This function returns a new Assignment
        object with the simplified expression.
        """
        return type(self)(
            name=self.name,
            value=self.value,
            components=self.components,
            unit_str=self.unit_str,
            unit=self.unit,
            expr=self.expr.simplify(),
            description=self.description,
            symbol=self.symbol,
            comment=self.comment,
        )


@attr.s(frozen=True, kw_only=True, slots=True)
class Intermediate(Assignment):
    """Intermediate is a type of Assignment that is not
    a StateDerivative"""


@attr.s(frozen=True, kw_only=True, slots=True)
class StateDerivative(Assignment):
    """A StateDerivative is an Assignment of the form
    `dX_dt = value` where X is a state. A StatedDerivative
    also holds a pointer to the State"""

    state: State = attr.ib()

    def resolve_expression(self, symbols: dict[str, sp.Symbol]) -> Assignment:
        if self.value is None:
            raise exceptions.ResolveExpressionError(name=self.name)
        expr = self.value.resolve(symbols)
        return StateDerivative(
            name=self.name,
            value=self.value,
            components=self.components,
            unit_str=self.unit_str,
            unit=self.unit,
            symbol=self.symbol,
            expr=expr,
            state=self.state,
            description=self.description,
            comment=self.comment,
        )



import sys
from unittest import mock

import pytest
from gotranx.schemes import get_scheme
from gotranx.codegen import PythonCodeGenerator
from gotranx.codegen import RHSArgument
from gotranx.ode import make_ode


@pytest.fixture(scope="module")
def ode(trans, parser):
    expr = """
    parameters(a=0)
    parameters("My component",
    sigma=ScalarParam(12.0, description="Some description"),
    rho=21.0,
    beta=2.4
    )
    states("My component", x=1.0, y=2.0,z=3.05)

    expressions("My component")
    rhoz = rho - z
    dy_dt = x*rhoz - y # millivolt
    dx_dt = sigma*(-x + y)
    betaz = beta*z
    dz_dt = -betaz + x*y
    """
    tree = parser.parse(expr)
    return make_ode(*trans.transform(tree), name="lorentz")


@pytest.fixture(scope="module")
def codegen(ode) -> PythonCodeGenerator:
    return PythonCodeGenerator(ode)


@pytest.fixture(scope="module")
def codegen_no_formatter(ode) -> PythonCodeGenerator:
    with mock.patch.dict(sys.modules, {"black": None, "ruff": None}):
        return PythonCodeGenerator(ode)


def test_python_codegen_state_index_no_formatter(codegen: PythonCodeGenerator):
    assert codegen.state_index() == (
        'state = {"x": 0, "z": 1, "y": 2}'
        "\n"
        "\n"
        "\ndef state_index(name: str) -> int:"
        '\n    """Return the index of the state with the given name'
        "\n"
        "\n    Arguments"
        "\n    ---------"
        "\n    name : str"
        "\n        The name of the state"
        "\n"
        "\n    Returns"
        "\n    -------"
        "\n    int"
        "\n        The index of the state"
        "\n"
        "\n    Raises"
        "\n    ------"
        "\n    KeyError"
        "\n        If the name is not a valid state"
        '\n    """'
        "\n"
        "\n    return state[name]"
        "\n"
    )


def test_python_codegen_initial_state_values(codegen: PythonCodeGenerator):
    assert codegen.initial_state_values() == (
        "def init_state_values(**values):"
        '\n    """Initialize state values"""'
        "\n    # x=1.0, z=3.05, y=2.0"
        "\n"
        "\n    states = numpy.array([1.0, 3.05, 2.0], dtype=numpy.float64)"
        "\n"
        "\n    for key, value in values.items():"
        "\n        states[state_index(key)] = value"
        "\n"
        "\n    return states"
        "\n"
    )


def test_python_codegen_parameter_index_no_formatter(codegen_no_formatter: PythonCodeGenerator):
    assert codegen_no_formatter.parameter_index() == (
        "\nparameter = {'a': 0, 'beta': 1, 'rho': 2, 'sigma': 3}"
        "\n"
        "\n"
        "\ndef parameter_index(name: str) -> int:"
        '\n    """Return the index of the parameter with the given name'
        "\n"
        "\n    Arguments"
        "\n    ---------"
        "\n    name : str"
        "\n        The name of the parameter"
        "\n"
        "\n    Returns"
        "\n    -------"
        "\n    int"
        "\n        The index of the parameter"
        "\n"
        "\n    Raises"
        "\n    ------"
        "\n    KeyError"
        "\n        If the name is not a valid parameter"
        '\n    """'
        "\n"
        "\n    return parameter[name]"
        "\n"
    )


def test_python_codegen_parameter_index(codegen: PythonCodeGenerator):
    assert codegen.parameter_index() == (
        'parameter = {"a": 0, "beta": 1, "rho": 2, "sigma": 3}'
        "\n"
        "\n"
        "\ndef parameter_index(name: str) -> int:"
        '\n    """Return the index of the parameter with the given name'
        "\n"
        "\n    Arguments"
        "\n    ---------"
        "\n    name : str"
        "\n        The name of the parameter"
        "\n"
        "\n    Returns"
        "\n    -------"
        "\n    int"
        "\n        The index of the parameter"
        "\n"
        "\n    Raises"
        "\n    ------"
        "\n    KeyError"
        "\n        If the name is not a valid parameter"
        '\n    """'
        "\n"
        "\n    return parameter[name]"
        "\n"
    )


def test_python_codegen_initial_parameter_values(codegen: PythonCodeGenerator):
    assert codegen.initial_parameter_values() == (
        "def init_parameter_values(**values):"
        '\n    """Initialize parameter values"""'
        "\n    # a=0, beta=2.4, rho=21.0, sigma=12.0"
        "\n"
        "\n    parameters = numpy.array([0, 2.4, 21.0, 12.0], dtype=numpy.float64)"
        "\n"
        "\n    for key, value in values.items():"
        "\n        parameters[parameter_index(key)] = value"
        "\n"
        "\n    return parameters"
        "\n"
    )


def test_python_codegen_missing_index_is_empty(codegen: PythonCodeGenerator):
    assert codegen.missing_index() == ""


@pytest.mark.parametrize(
    "order, arguments",
    [
        (
            RHSArgument.stp,
            ("states, t, parameters"),
        ),
        (
            RHSArgument.spt,
            ("states, parameters, t"),
        ),
        (
            RHSArgument.tsp,
            ("t, states, parameters"),
        ),
    ],
)
def test_python_codegen_rhs(order: str, arguments: str, codegen: PythonCodeGenerator):
    assert codegen.rhs(order=order) == (
        f"def rhs({arguments}):"
        "\n"
        "\n    # Assign states"
        "\n    x = states[0]"
        "\n    z = states[1]"
        "\n    y = states[2]"
        "\n"
        "\n    # Assign parameters"
        "\n    a = parameters[0]"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    betaz = beta * z"
        "\n    rhoz = rho - z"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    values[0] = dx_dt"
        "\n    dz_dt = -betaz + x * y"
        "\n    values[1] = dz_dt"
        "\n    dy_dt = rhoz * x - y"
        "\n    values[2] = dy_dt"
        "\n"
        "\n    return values"
        "\n"
    )


def test_python_codegen_forward_explicit_euler(codegen: PythonCodeGenerator):
    assert codegen.scheme(get_scheme("forward_explicit_euler")) == (
        "def forward_explicit_euler(states, t, dt, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    x = states[0]"
        "\n    z = states[1]"
        "\n    y = states[2]"
        "\n"
        "\n    # Assign parameters"
        "\n    a = parameters[0]"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    betaz = beta * z"
        "\n    rhoz = rho - z"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    values[0] = dt * dx_dt + x"
        "\n    dz_dt = -betaz + x * y"
        "\n    values[1] = dt * dz_dt + z"
        "\n    dy_dt = rhoz * x - y"
        "\n    values[2] = dt * dy_dt + y"
        "\n"
        "\n    return values"
        "\n"
    )


def test_python_codegen_forward_generalized_rush_larsen(codegen: PythonCodeGenerator):
    assert codegen.scheme(get_scheme("forward_generalized_rush_larsen")) == (
        "def forward_generalized_rush_larsen(states, t, dt, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    x = states[0]"
        "\n    z = states[1]"
        "\n    y = states[2]"
        "\n"
        "\n    # Assign parameters"
        "\n    a = parameters[0]"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    betaz = beta * z"
        "\n    rhoz = rho - z"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    dx_dt_linearized = -sigma"
        "\n    values[0] = x + numpy.where("
        "\n        numpy.logical_or((dx_dt_linearized > 1e-08), (dx_dt_linearized < -1e-08)),"
        "\n        dx_dt * (numpy.exp(dt * dx_dt_linearized) - 1) / dx_dt_linearized,"
        "\n        dt * dx_dt,"
        "\n    )"
        "\n    dz_dt = -betaz + x * y"
        "\n    values[1] = dt * dz_dt + z"
        "\n    dy_dt = rhoz * x - y"
        "\n    dy_dt_linearized = -1"
        "\n    values[2] = y + numpy.where("
        "\n        numpy.logical_or((dy_dt_linearized > 1e-08), (dy_dt_linearized < -1e-08)),"
        "\n        dy_dt * (numpy.exp(dt * dy_dt_linearized) - 1) / dy_dt_linearized,"
        "\n        dt * dy_dt,"
        "\n    )"
        "\n"
        "\n    return values"
        "\n"
    )


def test_python_conditional_expression(parser, trans):
    expr = """
    states(v=0)
    ah = Conditional(Ge(v, -40), 0, 0.057*exp(-(v + 80)/6.8))
    dv_dt = 0
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    ode = make_ode(*result, name="conditional")
    codegen = PythonCodeGenerator(ode)
    assert codegen.rhs() == (
        "def rhs(t, states, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    v = states[0]"
        "\n"
        "\n    # Assign parameters"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    dv_dt = 0"
        "\n    values[0] = dv_dt"
        "\n    ah = numpy.where((v >= -40), 0, 0.057 * numpy.exp((-(v + 80)) / 6.8))"
        "\n"
        "\n    return values"
        "\n"
    )


def test_python_exponential_with_power(parser, trans):
    expr = """
    states(v=0)
    tm = 0.06487*exp(-((v - 4.823)/51.12)**2)
    dv_dt = 0
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    ode = make_ode(*result, name="power")
    codegen = PythonCodeGenerator(ode)

    assert codegen.rhs() == (
        "def rhs(t, states, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    v = states[0]"
        "\n"
        "\n    # Assign parameters"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    dv_dt = 0"
        "\n    values[0] = dv_dt"
        "\n    tm = 0.06487 * numpy.exp(-(((v - 4.823) / 51.12) ** 2))"
        "\n"
        "\n    return values"
        "\n"
    )


def test_python_remove_unused_rhs(ode_unused):
    codegen_orig = PythonCodeGenerator(ode_unused)
    assert codegen_orig.rhs() == (
        "def rhs(t, states, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    unused_state = states[0]"
        "\n    x = states[1]"
        "\n    y = states[2]"
        "\n    z = states[3]"
        "\n"
        "\n    # Assign parameters"
        "\n    a = parameters[0]"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n    unused_parameter = parameters[4]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    unused_expression = 0"
        "\n    dunused_state_dt = 0"
        "\n    values[0] = dunused_state_dt"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    values[1] = dx_dt"
        "\n    dy_dt = x * (rho - z) - y"
        "\n    values[2] = dy_dt"
        "\n    dz_dt = (-beta) * z + x * y"
        "\n    values[3] = dz_dt"
        "\n"
        "\n    return values"
        "\n"
    )
    codegen_remove = PythonCodeGenerator(ode_unused, remove_unused=True)
    assert codegen_remove.rhs() == (
        "def rhs(t, states, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    x = states[1]"
        "\n    y = states[2]"
        "\n    z = states[3]"
        "\n"
        "\n    # Assign parameters"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    dunused_state_dt = 0"
        "\n    values[0] = dunused_state_dt"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    values[1] = dx_dt"
        "\n    dy_dt = x * (rho - z) - y"
        "\n    values[2] = dy_dt"
        "\n    dz_dt = (-beta) * z + x * y"
        "\n    values[3] = dz_dt"
        "\n"
        "\n    return values"
        "\n"
    )


def test_python_remove_unused_forward_explicit_euler(ode_unused):
    codegen_orig = PythonCodeGenerator(ode_unused)
    assert codegen_orig.scheme(get_scheme("forward_explicit_euler")) == (
        "def forward_explicit_euler(states, t, dt, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    unused_state = states[0]"
        "\n    x = states[1]"
        "\n    y = states[2]"
        "\n    z = states[3]"
        "\n"
        "\n    # Assign parameters"
        "\n    a = parameters[0]"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n    unused_parameter = parameters[4]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    unused_expression = 0"
        "\n    dunused_state_dt = 0"
        "\n    values[0] = dt * dunused_state_dt + unused_state"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    values[1] = dt * dx_dt + x"
        "\n    dy_dt = x * (rho - z) - y"
        "\n    values[2] = dt * dy_dt + y"
        "\n    dz_dt = (-beta) * z + x * y"
        "\n    values[3] = dt * dz_dt + z"
        "\n"
        "\n    return values"
        "\n"
    )
    codegen_remove = PythonCodeGenerator(ode_unused, remove_unused=True)
    assert codegen_remove.scheme(get_scheme("forward_explicit_euler")) == (
        "def forward_explicit_euler(states, t, dt, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    unused_state = states[0]"
        "\n    x = states[1]"
        "\n    y = states[2]"
        "\n    z = states[3]"
        "\n"
        "\n    # Assign parameters"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    dunused_state_dt = 0"
        "\n    values[0] = dt * dunused_state_dt + unused_state"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    values[1] = dt * dx_dt + x"
        "\n    dy_dt = x * (rho - z) - y"
        "\n    values[2] = dt * dy_dt + y"
        "\n    dz_dt = (-beta) * z + x * y"
        "\n    values[3] = dt * dz_dt + z"
        "\n"
        "\n    return values"
        "\n"
    )


def test_python_remove_unused_forward_generalized_rush_larsen(ode_unused):
    codegen_orig = PythonCodeGenerator(ode_unused)
    assert codegen_orig.scheme(get_scheme("forward_generalized_rush_larsen")) == (
        "def forward_generalized_rush_larsen(states, t, dt, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    unused_state = states[0]"
        "\n    x = states[1]"
        "\n    y = states[2]"
        "\n    z = states[3]"
        "\n"
        "\n    # Assign parameters"
        "\n    a = parameters[0]"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n    unused_parameter = parameters[4]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    unused_expression = 0"
        "\n    dunused_state_dt = 0"
        "\n    values[0] = dt * dunused_state_dt + unused_state"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    dx_dt_linearized = -sigma"
        "\n    values[1] = x + numpy.where("
        "\n        numpy.logical_or((dx_dt_linearized > 1e-08), (dx_dt_linearized < -1e-08)),"
        "\n        dx_dt * (numpy.exp(dt * dx_dt_linearized) - 1) / dx_dt_linearized,"
        "\n        dt * dx_dt,"
        "\n    )"
        "\n    dy_dt = x * (rho - z) - y"
        "\n    dy_dt_linearized = -1"
        "\n    values[2] = y + numpy.where("
        "\n        numpy.logical_or((dy_dt_linearized > 1e-08), (dy_dt_linearized < -1e-08)),"
        "\n        dy_dt * (numpy.exp(dt * dy_dt_linearized) - 1) / dy_dt_linearized,"
        "\n        dt * dy_dt,"
        "\n    )"
        "\n    dz_dt = (-beta) * z + x * y"
        "\n    dz_dt_linearized = -beta"
        "\n    values[3] = z + numpy.where("
        "\n        numpy.logical_or((dz_dt_linearized > 1e-08), (dz_dt_linearized < -1e-08)),"
        "\n        dz_dt * (numpy.exp(dt * dz_dt_linearized) - 1) / dz_dt_linearized,"
        "\n        dt * dz_dt,"
        "\n    )"
        "\n"
        "\n    return values"
        "\n"
    )
    codegen_remove = PythonCodeGenerator(ode_unused, remove_unused=True)
    assert codegen_remove.scheme(get_scheme("forward_generalized_rush_larsen")) == (
        "def forward_generalized_rush_larsen(states, t, dt, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    unused_state = states[0]"
        "\n    x = states[1]"
        "\n    y = states[2]"
        "\n    z = states[3]"
        "\n"
        "\n    # Assign parameters"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n"
        "\n    # Assign expressions"
        "\n"
        "\n    values = numpy.zeros_like(states, dtype=numpy.float64)"
        "\n    dunused_state_dt = 0"
        "\n    values[0] = dt * dunused_state_dt + unused_state"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    dx_dt_linearized = -sigma"
        "\n    values[1] = x + numpy.where("
        "\n        numpy.logical_or((dx_dt_linearized > 1e-08), (dx_dt_linearized < -1e-08)),"
        "\n        dx_dt * (numpy.exp(dt * dx_dt_linearized) - 1) / dx_dt_linearized,"
        "\n        dt * dx_dt,"
        "\n    )"
        "\n    dy_dt = x * (rho - z) - y"
        "\n    dy_dt_linearized = -1"
        "\n    values[2] = y + numpy.where("
        "\n        numpy.logical_or((dy_dt_linearized > 1e-08), (dy_dt_linearized < -1e-08)),"
        "\n        dy_dt * (numpy.exp(dt * dy_dt_linearized) - 1) / dy_dt_linearized,"
        "\n        dt * dy_dt,"
        "\n    )"
        "\n    dz_dt = (-beta) * z + x * y"
        "\n    dz_dt_linearized = -beta"
        "\n    values[3] = z + numpy.where("
        "\n        numpy.logical_or((dz_dt_linearized > 1e-08), (dz_dt_linearized < -1e-08)),"
        "\n        dz_dt * (numpy.exp(dt * dz_dt_linearized) - 1) / dz_dt_linearized,"
        "\n        dt * dz_dt,"
        "\n    )"
        "\n"
        "\n    return values"
        "\n"
    )


def test_python_monitored(codegen: PythonCodeGenerator):
    assert codegen.monitor_values() == (
        "def monitor_values(t, states, parameters):"
        "\n"
        "\n    # Assign states"
        "\n    x = states[0]"
        "\n    z = states[1]"
        "\n    y = states[2]"
        "\n"
        "\n    # Assign parameters"
        "\n    a = parameters[0]"
        "\n    beta = parameters[1]"
        "\n    rho = parameters[2]"
        "\n    sigma = parameters[3]"
        "\n"
        "\n    # Assign expressions"
        "\n    shape = 5 if len(states.shape) == 1 else (5, states.shape[1])"
        "\n    values = numpy.zeros(shape)"
        "\n    betaz = beta * z"
        "\n    values[0] = betaz"
        "\n    rhoz = rho - z"
        "\n    values[1] = rhoz"
        "\n    dx_dt = sigma * (-x + y)"
        "\n    values[2] = dx_dt"
        "\n    dz_dt = -betaz + x * y"
        "\n    values[3] = dz_dt"
        "\n    dy_dt = rhoz * x - y"
        "\n    values[4] = dy_dt"
        "\n"
        "\n    return values"
        "\n"
    )


def test_python_monitored_index(codegen: PythonCodeGenerator):
    assert codegen.monitor_index() == (
        'monitor = {"betaz": 0, "rhoz": 1, "dx_dt": 2, "dz_dt": 3, "dy_dt": 4}'
        "\n"
        "\n"
        "\ndef monitor_index(name: str) -> int:"
        '\n    """Return the index of the monitor with the given name'
        "\n"
        "\n    Arguments"
        "\n    ---------"
        "\n    name : str"
        "\n        The name of the monitor"
        "\n"
        "\n    Returns"
        "\n    -------"
        "\n    int"
        "\n        The index of the monitor"
        "\n"
        "\n    Raises"
        "\n    ------"
        "\n    KeyError"
        "\n        If the name is not a valid monitor"
        '\n    """'
        "\n"
        "\n    return monitor[name]"
        "\n"
    )



from pathlib import Path
import gotranx
import pytest

here = Path(__file__).parent

cellml_file = here / "cellml_files" / "ToRORd_dynCl_mid.cellml"
ode_file = here / "cellml_files" / "ToRORd_dynCl_mid.ode"


@pytest.fixture(scope="module", autouse=True)
def ode():
    _ode = gotranx.myokit.cellml_to_gotran(cellml_file)
    _ode.save(ode_file)
    yield gotranx.load_ode(ode_file)
    ode_file.unlink()


@pytest.mark.benchmark
def test_cell2gotran():
    gotranx.myokit.cellml_to_gotran(cellml_file)


@pytest.mark.benchmark
def test_save(ode, tmp_path):
    ode.save(tmp_path / "test.ode")


@pytest.mark.benchmark
def test_python_no_schemes(ode):
    gotranx.cli.gotran2py.get_code(ode)


@pytest.mark.benchmark
def test_python_explicit_euler(ode):
    gotranx.cli.gotran2py.get_code(ode, scheme=[gotranx.schemes.Scheme.explicit_euler])


@pytest.mark.benchmark
def test_python_generalized_rush_larsen(ode):
    gotranx.cli.gotran2py.get_code(ode, scheme=[gotranx.schemes.Scheme.generalized_rush_larsen])


@pytest.mark.benchmark
def test_c_no_schemes(ode):
    gotranx.cli.gotran2c.get_code(ode)


@pytest.mark.benchmark
def test_c_explicit_euler(ode):
    gotranx.cli.gotran2c.get_code(ode, scheme=[gotranx.schemes.Scheme.explicit_euler])


@pytest.mark.benchmark
def test_c_generalized_rush_larsen(ode):
    gotranx.cli.gotran2c.get_code(ode, scheme=[gotranx.schemes.Scheme.generalized_rush_larsen])



from __future__ import annotations

from collections import defaultdict
from typing import TypeVar, Callable
from functools import reduce
from sympy.printing.str import StrPrinter
from structlog import get_logger

from .. import atoms
from ..ode import ODE
from .base import _print_Piecewise

T = TypeVar("T")
logger = get_logger()


def break_comment_at_80(acc, x):
    if x == "#":
        return acc + "\n#"

    if len(acc.split("\n")[-1]) + len(x) > 80:
        return acc + "\n# " + x
    else:
        return acc + " " + x


class BaseGotranODECodePrinter(StrPrinter):
    def _print_Relational(self, expr):
        # v = super()._print_Relational(expr)
        lhs = self._print(expr.lhs)
        rhs = self._print(expr.rhs)

        relop2str = {
            "<": "Lt",
            "<=": "Le",
            ">": "Gt",
            ">=": "Ge",
            "==": "Eq",
            "!=": "Ne",
        }
        relop = relop2str[expr.rel_op]
        return f"{relop}({lhs}, {rhs})"

    def _print_Or(self, expr):
        return f"Or({', '.join(self._print(a) for a in expr.args)})"

    def _print_And(self, expr):
        return f"And({', '.join(self._print(a) for a in expr.args)})"

    def _print_BooleanFalse(self, expr):
        return "0"

    def _print_BooleanTrue(self, expr):
        return "1"

    def _print_Piecewise(self, expr):
        conds, exprs = _print_Piecewise(self, expr)

        result = []

        for c, e in zip(conds, exprs):
            result.append("Conditional(")
            result.append(f"{c}")
            result.append(", ")
            result.append(f"{e}")
            result.append(", ")

        if c == "1":
            result = result[:-6]
            result.append(f", {e}")
        else:
            raise ValueError("Last condition in Piecewise must be True")

        result.append(")" * (len(conds) - 1))
        return "".join(result)


class GotranODECodePrinter(BaseGotranODECodePrinter):
    def __init__(self, ode: ODE, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.ode = ode

    def print_comments(self) -> str:
        if len(self.ode.comments) == 0:
            return ""
        text_lst = []

        for comment in self.ode.comments:
            text_lst.append(reduce(break_comment_at_80, comment.text.split(" ")))

        text = "\n".join(text_lst)
        if len(text) > 0:
            text = "# " + "\n# ".join(text.strip().split("\n"))
            text += "\n\n"
        return text

    def print_states(self) -> str:
        d: dict[tuple[str, ...], list[atoms.State]] = defaultdict(list)
        for state in self.ode.states:
            d[state.components].append(state)

        text = ""
        for components, states in d.items():
            text += start_odeblock("states", names=components) + "\n"
            text += ",\n".join([print_ScalarParam(s, doprint=self.doprint) for s in states])
            text += "\n)\n\n"
        return text

    def print_parameters(self) -> str:
        d: dict[tuple[str, ...], list[atoms.Parameter]] = defaultdict(list)
        for parameter in self.ode.parameters:
            d[parameter.components].append(parameter)

        text = ""
        for components, parameters in d.items():
            text += start_odeblock("parameters", names=components) + "\n"
            text += ",\n".join([print_ScalarParam(p, doprint=self.doprint) for p in parameters])
            text += "\n)\n\n"

        return text

    def print_assignments(self) -> str:
        d: dict[tuple[str, ...], list[atoms.Assignment]] = defaultdict(list)
        for i in self.ode.intermediates + self.ode.state_derivatives:
            d[i.components].append(i)

        text = ""
        for components, intermediates in d.items():
            text += start_odeblock("expressions", names=components, is_expression=True) + "\n"
            text += "\n".join([print_assignment(i, doprint=self.doprint) for i in intermediates])
            text += "\n\n"

        return text


def start_odeblock(case: str, names: tuple[str, ...] = (), is_expression: bool = False):
    if len(names) == 0 or (len(names) == 1 and names[0] == ""):
        if is_expression:
            return ""
        return f"{case}("
    else:
        args = ", ".join(f'"{n}"' for n in names)

        if is_expression:
            return f"{case}({args})"
        else:
            return f"{case}({args},"


def join(*args: str) -> str:
    non_empty_args = [a for a in args if a != ""]
    if len(non_empty_args) == 0:
        return ""
    elif len(non_empty_args) == 1:
        return ", " + non_empty_args[0]
    else:
        return ", " + ", ".join(args)


def print_ScalarParam(p: atoms.Atom, doprint: Callable[[str], str]) -> str:
    unit_str = "" if p.unit_str is None else f'unit="{p.unit_str}"'
    description = "" if p.description is None else f'description="{p.description}"'
    if unit_str == "" and description == "":
        ret = f"{p.name}={doprint(p.value)}"  # type: ignore
    else:
        kwargs_str = join(unit_str, description)
        ret = f"{p.name}=ScalarParam({doprint(p.value)}{kwargs_str})"  # type: ignore

    logger.debug(f"Saving parameters {p} as {ret!r}")
    return ret


def print_assignment(a: atoms.Assignment, doprint: Callable[[str], str]) -> str:
    s = f"{a.name} = {doprint(a.expr)}"
    unit_or_comment = ""
    if a.comment is not None:
        unit_or_comment = a.comment.text
    if a.unit_str is not None and a.unit_str != "1":
        unit_or_comment = a.unit_str
    if unit_or_comment != "":
        s += f" # {unit_or_comment}"

    return s
